import os
import shutil
from collections.abc import Iterator
from pathlib import Path
from typing import Any
from unittest.mock import AsyncMock, Mock, patch

import pytest
from aiohttp import ClientResponseError
from huggingface_hub import AsyncInferenceClient
from transformers import PreTrainedTokenizerBase

from artkit.model.llm import (
    CachedChatModel,
    CachedCompletionModel,
    ChatFromCompletionModel,
)
from artkit.model.llm.base import ChatModel, ChatModelConnector, CompletionModel
from artkit.model.llm.huggingface import HuggingfaceChat, HuggingfaceCompletion
from artkit.model.util import RateLimitException

_ = pytest.importorskip("huggingface_hub")

#######################################################################################
#                                     Constants                                       #
#######################################################################################

MODEL_MISTRAL_8x7B = "mistralai/Mixtral-8x7B-Instruct-v0.1"
MODEL_MISTRAL_7B = "mistralai/Mistral-7B-v0.1"
MODEL_GPT2 = "openai-community/gpt2"
ZEPHYR_7B_MODEL = "HuggingFaceH4/zephyr-7b-beta"
EXAMPLE_URL = "http://huggingface.com"

PROMPT = "In one and only one word, the color of the sky is:"
PROMPT_COMPLETION = "blue"

#######################################################################################
#                                     UTILs                                           #
#######################################################################################


def _identity(x: Any, *_args: Any, **_kwargs: Any) -> Any:
    return x


async def _async_identity(x: Any, *_args: Any, **_kwargs: Any) -> Any:
    return x


@pytest.mark.asyncio
async def test_huggingface_chat(
    cached_huggingface_completion_chat: ChatFromCompletionModel[Any],
) -> None:
    """
    Test the HuggingfaceLLM class.
    """
    assert (
        await cached_huggingface_completion_chat.with_system_prompt(
            "Your job is to answer a quiz question in a single word"
        ).get_response(message="What color is the sky?")
    )[0].strip().lower() == PROMPT_COMPLETION

    # check that [/AGENT] gets clipped
    assert cached_huggingface_completion_chat.postprocess_response("test [/AGENT]") == [
        "test"
    ]


@pytest.mark.asyncio
async def test_huggingface_async(cached_huggingface_async: CompletionModel) -> None:
    """
    Test the AsyncHuggingfaceLLM class.
    """
    assert (
        await cached_huggingface_async.get_completion(prompt=PROMPT)
    ).strip().lower() == PROMPT_COMPLETION


@pytest.mark.asyncio
async def test_cached_huggingface_chat_async(
    cached_huggingface_chat_async: ChatModel,
) -> None:
    """
    Test the HuggingfaceChat class.
    """
    assert (
        await cached_huggingface_chat_async.with_system_prompt(
            "Your job is to answer a quiz question with a single word, "
            "lowercase, with no punctuation"
        ).get_response(message="What color is the sky?")
    )[0] == "blue"


@pytest.mark.asyncio
async def test_huggingface_chat_async(
    hugging_face_chat: ChatModelConnector[AsyncInferenceClient],
    zephyr_7b_tokenizer: PreTrainedTokenizerBase,
) -> None:
    with patch("aiohttp.ClientSession") as MockClientSession:

        # Mock the response object
        mock_post = Mock()
        mock_post.read = AsyncMock(return_value=b'[{"generated_text": "blue"}]')
        mock_post.return_value.status = 200

        # Set up the mock connection object
        mock_connection = AsyncMock()
        mock_connection.post.return_value = mock_post
        MockClientSession.return_value = mock_connection

        with patch(
            "transformers.AutoTokenizer.from_pretrained"
        ) as mock_from_pretrained:
            mock_from_pretrained.return_value = zephyr_7b_tokenizer
            assert (
                await hugging_face_chat.with_system_prompt(
                    "Your job is to answer a quiz question with a single word, "
                    "lowercase, with no punctuation"
                ).get_response(message="What color is the sky?")
            )[0] == "blue"


@pytest.mark.asyncio
async def test_huggingface_retry(
    hugging_face_chat: HuggingfaceChat,
    zephyr_7b_tokenizer: PreTrainedTokenizerBase,
    caplog: pytest.LogCaptureFixture,
) -> None:
    with patch("aiohttp.ClientSession") as MockClientSession:

        # Mock the response object
        mock_post = AsyncMock()
        mock_post.read.return_value = b'[{"generated_text": "blue"}]'
        mock_post.json.return_value = {"error": "Rate limit exceeded"}
        mock_post.return_value.status = 429

        # Request info mock
        request_info = AsyncMock()
        request_info.real_url = (
            "https://api-inference.huggingface.co/models/HuggingFaceH4/zephyr-7b-beta"
        )

        def f() -> None:
            err = ClientResponseError(
                request_info=request_info,
                history=AsyncMock(),
                status=429,
                message="Rate limit exceeded",
            )
            raise err

        mock_post.raise_for_status = f

        # Set up the mock connection object
        mock_connection = AsyncMock()
        mock_connection.post.return_value = mock_post
        MockClientSession.return_value = mock_connection

        # Get number of awaited retries
        n_retries = hugging_face_chat.max_retries

        with patch(
            "transformers.AutoTokenizer.from_pretrained"
        ) as mock_from_pretrained:
            mock_from_pretrained.return_value = zephyr_7b_tokenizer
            with pytest.raises(RateLimitException):
                await hugging_face_chat.with_system_prompt(
                    "Your job is to answer a quiz question with a single word, "
                    "lowercase, with no punctuation"
                ).get_response(message="What color is the sky?")
            assert mock_connection.post.call_count == n_retries

        assert (
            len(
                [
                    record
                    for record in caplog.records
                    if record.message.startswith("Rate limit exceeded")
                ]
            )
            == n_retries
        )


def test_error_on_missing_chat_template(
    zephyr_7b_tokenizer: PreTrainedTokenizerBase, hf_token: str
) -> None:
    with pytest.raises(ValueError):
        with patch(
            "transformers.AutoTokenizer.from_pretrained"
        ) as mock_from_pretrained:
            zephyr_7b_tokenizer.chat_template = None
            mock_from_pretrained.return_value = zephyr_7b_tokenizer
            HuggingfaceChat(
                max_new_tokens=1,
                temperature=1.0,
                api_key_env=hf_token,
                model_id=ZEPHYR_7B_MODEL,
            ).get_tokenizer()


@pytest.mark.asyncio
async def test_custom_chat_template(
    zephyr_7b_tokenizer: PreTrainedTokenizerBase, hf_token: str
) -> None:
    with patch("transformers.AutoTokenizer.from_pretrained") as mock_from_pretrained:
        mock_from_pretrained.return_value = zephyr_7b_tokenizer
        assert zephyr_7b_tokenizer.chat_template is not None
        llm = HuggingfaceChat(
            max_tokens=1,
            temperature=1.0,
            chat_template=(
                "{{ bos_token }}"
                "{% for message in messages %}"
                "{% if (message['role'] == 'user') != (loop.index0 % 2 == 0) %}"
                "{{ raise_exception("
                "'Conversation roles must alternate user/assistant/user/assistant/...'"
                ") }}"
                "{% endif %}"
                "{% if message['role'] == 'user' %}"
                "{{ '[INST] ' + message['content'] + ' [/INST]' }}"
                "{% elif message['role'] == 'assistant' %}"
                "{{ message['content'] + eos_token}}"
                "{% else %}"
                "{{ raise_exception('Only user and assistant roles are supported!') }}"
                "{% endif %}"
                "{% endfor %}"
            ),
            chat_template_params={"bos_token": "<s>"},
            api_key_env=hf_token,
            model_id=ZEPHYR_7B_MODEL,
        )

        mock_client = AsyncMock()
        mock_client.text_generation = _async_identity
        with patch.object(llm, "get_client") as mock_get_client:
            mock_get_client.return_value = mock_client
            assert (
                await llm._get_client_response([{"role": "user", "content": PROMPT}])
            )[0] == f"<s>[INST] {PROMPT} [/INST]"


#######################################################################################
#                                     FIXTURES                                        #
#######################################################################################


@pytest.fixture
def hf_token() -> str:
    api_key = "HF_TOKEN"
    # create a dummy key if not found
    if api_key not in os.environ:
        os.environ[api_key] = "<HF API key>"
    return api_key


@pytest.fixture
def cached_huggingface_async(
    hf_token: str, data_path: Path
) -> Iterator[CompletionModel]:
    database = data_path / "_copy_huggingface_async.db"
    shutil.copyfile(data_path / "huggingface_async.db", database)

    llm = HuggingfaceCompletion(
        api_key_env=hf_token, model_id=MODEL_MISTRAL_8x7B, max_tokens=1, temperature=1.0
    )

    yield CachedCompletionModel(model=llm, database=database)

    os.remove(database)


@pytest.fixture
def hugging_face_chat(
    zephyr_7b_tokenizer: PreTrainedTokenizerBase, hf_token: str
) -> ChatModel:

    with patch("transformers.AutoTokenizer.from_pretrained") as mock_from_pretrained:
        mock_from_pretrained.return_value = zephyr_7b_tokenizer

        return HuggingfaceChat(
            max_new_tokens=1,
            temperature=1.0,
            api_key_env=hf_token,
            model_id=ZEPHYR_7B_MODEL,
            initial_delay=0.1,
            exponential_base=1.5,
            max_retries=2,
        )


@pytest.fixture
def cached_huggingface_chat_async(
    hugging_face_chat: ChatModel, data_path: Path
) -> Iterator[ChatModel]:
    database = data_path / "_copy_huggingface_chat_async.db"
    shutil.copyfile(data_path / "huggingface_chat_async.db", database)

    yield CachedChatModel(model=hugging_face_chat, database=database)

    os.remove(database)


@pytest.fixture
def cached_huggingface_completion_chat(
    cached_huggingface_async: CompletionModel,
) -> ChatFromCompletionModel[Any]:
    return ChatFromCompletionModel(model=cached_huggingface_async)


@pytest.fixture
def zephyr_7b_tokenizer() -> PreTrainedTokenizerBase:
    tokenizer = PreTrainedTokenizerBase(clean_up_tokenization_spaces=True)
    tokenizer.chat_template = (
        "{% for message in messages %}\n{% if message['role'] == 'user' %}\n"
        "{{ '<|user|>\n' + message['content'] + eos_token }}\n{% elif message['role'] == 'system' %}\n"
        "{{ '<|system|>\n' + message['content'] + eos_token }}\n{% elif message['role'] == 'assistant' %}\n"
        "{{ '<|assistant|>\n'  + message['content'] + eos_token }}\n{% endif %}\n"
        "{% if loop.last and add_generation_prompt %}\n{{ '<|assistant|>' }}\n{% endif %}\n{% endfor %}"
    )
    tokenizer.eos_token = "</s>"
    return tokenizer



import os
import shutil
from collections.abc import Iterator
from pathlib import Path
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from groq import RateLimitError

from artkit.model.llm import CachedChatModel
from artkit.model.llm.base import ChatModel
from artkit.model.llm.groq import GroqChat
from artkit.model.util import RateLimitException

groq = pytest.importorskip("groq")


@pytest.mark.asyncio
async def test_groq(groq_chat: GroqChat) -> None:
    # Mock groq Client
    with patch("artkit.model.llm.groq._groq.AsyncGroq") as mock_get_client:
        # Mock groq Client response
        mock_response = AsyncMock(
            return_value=AsyncMock(
                choices=[MagicMock(message=MagicMock(content="blue", role="assistant"))]
            )
        )

        # Set mock response as return value
        mock_get_client.return_value.chat.completions.create = mock_response

        # Call mocked model
        messages = await groq_chat.get_response(
            message="What color is the sky? Please answer in one word."
        )
        assert "blue" in messages[0].lower()


@pytest.mark.asyncio
async def test_cached_groq(
    cached_groq: ChatModel,
) -> None:
    messages = await cached_groq.get_response(
        message="What color is the sky? Please answer in one word."
    )
    assert "blue" in messages[0].lower()


@pytest.mark.asyncio
async def test_groq_retry(
    groq_chat: GroqChat, caplog: pytest.LogCaptureFixture
) -> None:
    with patch("artkit.model.llm.groq._groq.AsyncGroq") as MockClientSession:

        # Response mock
        response = AsyncMock()
        response.status_code = 429

        MockClientSession.return_value.chat.completions.create.side_effect = (
            RateLimitError(
                message="Rate Limit exceeded",
                response=response,
                body=AsyncMock(),
            )
        )

        with pytest.raises(RateLimitException):
            await groq_chat.with_system_prompt(
                "Your job is to answer a quiz question with a single word, "
                "lowercase, with no punctuation"
            ).get_response(message="What color is the sky?")
        assert MockClientSession.return_value.chat.completions.create.call_count == 2
    assert (
        len(
            [
                record
                for record in caplog.records
                if record.message.startswith("Rate limit exceeded")
            ]
        )
        == groq_chat.max_retries
    )


@pytest.fixture
def groq_chat() -> GroqChat:
    api_key_env = "TEST"
    os.environ[api_key_env] = "test"
    return GroqChat(
        max_tokens=10,
        temperature=1.0,
        model_id="mixtral-8x7b-32768",
        api_key_env=api_key_env,
        initial_delay=0.1,
        exponential_base=1.5,
        max_retries=2,
    )


@pytest.fixture
def cached_groq(data_path: Path, groq_chat: GroqChat) -> Iterator[ChatModel]:
    database = data_path / "_copy_groq.db"
    shutil.copyfile(data_path / "groq.db", database)

    yield CachedChatModel(model=groq_chat, database=database)

    os.remove(database)



# -----------------------------------------------------------------------------
# © 2024 Boston Consulting Group. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
Implementation of the Image class for I/O handling of Vision and Diffusion models.
"""

import base64
from typing import Any, overload

from pytools.api import inheritdoc
from pytools.expression import Expression, HasExpressionRepr
from pytools.expression.atomic import Id
from pytools.http import fetch_url

__all__ = [
    "Image",
]


#
# Classes
#


@inheritdoc(match="""[see superclass]""")
class Image(HasExpressionRepr):
    """Representation of an image for I/O handling of Vision and Diffusion models"""

    #: Binary data of image
    _data: bytes | None = None

    #: URL of image
    url: str | None = None

    @overload
    def __init__(self, *, data: bytes) -> None:
        """[see below]"""

    @overload
    def __init__(self, *, url: str) -> None:
        """[see below]"""

    def __init__(self, *, data: bytes | None = None, url: str | None = None) -> None:
        """
        :param data: the binary image data
        :param url: a URL to fetch the image from
        """
        if data is not None:
            if url is not None:
                raise ValueError("Exactly one of arg data and arg url must be provided")
            self._data = data
        elif url is not None:
            self.url = url
        else:
            raise ValueError("Either arg data or arg url must be provided")

    @property
    def data(self) -> bytes:
        """
        The binary image data.
        """
        data = self._data
        if data is None:
            # Fetch the image data from the URL
            self._data = data = self._fetch_image_from_url()

        return data

    def _fetch_image_from_url(self) -> bytes:
        """
        Download an image from URL and encodes it in binary format.

        :return: the binary image data
        """
        url = self.url
        assert url is not None, "URL must be set to fetch image from URL"

        # Send a GET request to the image URL
        return fetch_url(url)

    def to_b64(self) -> bytes:
        """
        Encode the image data into base64 format.

        :return: the base64-encoded image
        """
        return base64.b64encode(self.data)

    def to_b64_string(self) -> str:
        """
        Encode the image data as a base64 string.

        The main purpose of this method is to allow for JSON serialization.

        :return: a base64-encoded string representing the image
        """
        return base64.b64encode(self.data).decode("ascii")

    def show(
        self, width: int | None = None, height: int | None = None
    ) -> None:  # pragma: no cover
        """
        Display the image using IPython.

        :param width: the width of the image (optional)
        :param height: the height of the image (optional)
        """
        try:
            from IPython.display import Image, display

            display(Image(data=self.data, width=width, height=height))
        except ImportError as e:
            raise ImportError("IPython is required to display images") from e

    def __hash__(self) -> int:
        return hash((self.url, self.data))

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, Image)
            and self.url == other.url
            and self.data == other.data
        )

    def _repr_png_(self) -> bytes:
        """
        Render the image as a PNG for Jupyter notebooks.

        :return: the PNG data
        """
        return self.data

    def to_expression(self) -> Expression:
        """[see superclass]"""
        id = Id(type(self))
        if self.url is not None:
            return id(url=self.url)
        else:
            return id(data=self.data)



# -----------------------------------------------------------------------------
# © 2024 Boston Consulting Group. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
This module provides a functional API for defining and executing
complex and asynchronous workflows. The API is inspired by the functional programming
paradigm and is designed to be simple, expressive, and composable.

It is based on the :mod:`fluxus.functional` module of the *fluxus* library.
Please refer to the *fluxus* documentation for more information.

The functions :func:`step`, :func:`chain`, and :func:`parallel` are used to
define the steps of a flow. Function :func:`step` defines a single step, function
:func:`chain` composes steps or flows sequentially, and the :func:`parallel` composes
steps or flows in parallel.

Alternatively, the ``>>`` operator can be used to chain steps or flows, and the
``&`` operator can be used to compose steps or flows in parallel.

Function :func:`passthrough` can be included in a parallel composition to pass the
unchanged input through to the next step.

Function :func:`run` is used to execute a flow. The flow is executed asynchronously,
and all results are returned as a :class:`RunResult` instance. Run results include
the input to the flow as well as all intermediate and final results along all paths
through the flow.

Example:

.. code-block:: python

    from artkit.flow import step, chain, parallel, passthrough, run
    from collections.abc import AsyncIterator
    from typing import Any

    def add_one(x) -> dict[str, Any]:
        # We can return a single dict
        return dict(x=x + 1)

    async def add_two(x) -> dict[str, Any]:
        # We can use asynchronous functions
        return dict(x=x + 2)

    async def add_three_or_five(x) -> AsyncIterator[dict[str, Any]]:
        # We can yield rather than return values to generate multiple products
        yield dict(x=x + 3)
        yield dict(x=x + 5)

    flow = chain(
        step("add_1", add_one),
        parallel(
            step("add_2", add_two),
            step("add_3_or_5", add_three_or_five),
            passthrough(),
        ),
    )

    result = run(flow, input=[dict(x=1), dict(x=10)])

    print(result)

This will output:

.. code-block:: python

    RunResult(
        [
            {'input': {'x': 1}, 'add_1': {'x': 2}, 'add_2': {'x': 4}},
            {'input': {'x': 10}, 'add_1': {'x': 11}, 'add_2': {'x': 13}}
        ],
        [
            {'input': {'x': 1}, 'add_1': {'x': 2}, 'add_3_or_5': {'x': 5}},
            {'input': {'x': 1}, 'add_1': {'x': 2}, 'add_3_or_5': {'x': 7}},
            {'input': {'x': 10}, 'add_1': {'x': 11}, 'add_3_or_5': {'x': 14}},
            {'input': {'x': 10}, 'add_1': {'x': 11}, 'add_3_or_5': {'x': 16}}
        ],
        [
            {'input': {'x': 1}, 'add_1': {'x': 2}},
            {'input': {'x': 10}, 'add_1': {'x': 11}}
        ]
    )

The following functions and classes are provided by this module:
"""


def __mirror_fluxus() -> None:  # pragma: no cover
    import os
    from typing import Any

    import fluxus.functional as fluxus

    # Get all public definitions from the fluxus module as tuples (name, obj)
    fluxus_definitions: list[tuple[str, Any]] = sorted(
        (name, obj)
        for name, obj in vars(fluxus).items()
        if not name.startswith("_") and callable(obj)
    )

    # Get the globals of this module
    globals_ = globals()

    # Generate bullet points for the module docstring
    docstring = "\n".join(
        f"- :obj:`~fluxus.functional.{name}`" for name, obj in fluxus_definitions
    )

    # Add the bullet points to the module docstring
    globals_["__doc__"] += "\n\n" + docstring

    # Check if we are building the Sphinx documentation
    if "SPHINX_BUILD" in os.environ:
        # we are building the documentation - we are done
        return

    # We are not building the Sphinx documentation - add the fluxus definitions to this
    # module

    # Add all public definitions from the fluxus module to this module
    for name, obj in fluxus_definitions:
        globals_[name] = obj

    # Set the __all__ attribute to all public symbols
    globals_["__all__"] = [name for name, _ in fluxus_definitions]


# Update the module attributes of the functions and classes imported from fluxus
__mirror_fluxus()

# Clean up the namespace
del __mirror_fluxus



import os
import shutil
from collections.abc import Iterator
from pathlib import Path
from typing import Any
from unittest.mock import AsyncMock, Mock, patch

import pytest
import torch
from transformers import PreTrainedTokenizerBase

from artkit.model.llm import CachedCompletionModel
from artkit.model.llm.base import ChatModel, CompletionModel
from artkit.model.llm.huggingface import (
    HuggingfaceChatLocal,
    HuggingfaceCompletionLocal,
)

_ = pytest.importorskip("huggingface_hub")

#######################################################################################
#                                     Constants                                       #
#######################################################################################

MODEL_MISTRAL_8x7B = "mistralai/Mixtral-8x7B-Instruct-v0.1"
MODEL_MISTRAL_7B = "mistralai/Mistral-7B-v0.1"
MODEL_GPT2 = "openai-community/gpt2"
ZEPHYR_7B_MODEL = "HuggingFaceH4/zephyr-7b-beta"
EXAMPLE_URL = "http://huggingface.com"

PROMPT = "In one and only one word, the color of the sky is:"
PROMPT_COMPLETION = "blue"

#######################################################################################
#                                     UTILs                                           #
#######################################################################################


@pytest.mark.asyncio
async def test_huggingface_local(cached_huggingface_local: CompletionModel) -> None:
    """
    Test the HuggingfaceLocalLLM class.
    """

    assert cached_huggingface_local.model_id == MODEL_MISTRAL_7B

    assert (
        await cached_huggingface_local.get_completion(prompt=PROMPT)
    ).strip().lower() == PROMPT_COMPLETION


@pytest.mark.asyncio
async def test_local_hf_completion(
    huggingface_local_completion: CompletionModel,
) -> None:
    with patch(
        "artkit.model.llm.huggingface._huggingface_local.HuggingfaceCompletionLocal.get_client"
    ) as mock_get_client:
        with patch(
            "artkit.model.llm.huggingface._huggingface_local.HuggingfaceCompletionLocal.get_tokenizer"
        ) as mock_get_tokenizer:
            _RESULT = "blue"
            _MAP = {1: "What color", 2: " is the", 3: " sky?", 4: _RESULT}
            _PROMPT = "".join([_MAP[i] for i in range(1, 4)])

            # Setup mock client behavior
            mock_client = AsyncMock()
            mock_client.device = "cpu"
            mock_client.generate = lambda tokens, *_args, **_kwargs: torch.cat(
                (tokens, torch.Tensor([[4]])), dim=1
            )
            mock_get_client.return_value = mock_client

            # Setup mock tokenizer behavior
            mock_tokenizer = Mock()
            mock_tokenizer.return_value.input_ids = torch.tensor([[1, 2, 3]])

            def mock_decode(tokens: torch.Tensor, *_args: Any, **_kwargs: Any) -> str:
                result = ""
                for token in tokens:
                    result += _MAP[token.item()]
                return result

            mock_tokenizer.decode = mock_decode
            mock_get_tokenizer.return_value = mock_tokenizer

            # Assert output being passed through correctly
            assert (
                await huggingface_local_completion.get_completion(prompt=_PROMPT)
            ) == _RESULT


@pytest.mark.asyncio
async def test_local_hf_chat(huggingface_local_chat: ChatModel) -> None:
    with patch(
        "artkit.model.llm.huggingface._huggingface_local.HuggingfaceChatLocal.get_client"
    ) as mock_get_client:
        with patch(
            "artkit.model.llm.huggingface._huggingface_local.HuggingfaceChatLocal.get_tokenizer"
        ) as mock_get_tokenizer:
            _PROMPT = "What color is the sky?"
            _RESULT = "blue"

            # Setup mock client behavior
            mock_client = AsyncMock()
            mock_client.device = "cpu"
            mock_client.generate = lambda tokens, *_args, **_kwargs: torch.cat(
                (tokens, torch.Tensor([[4]])), dim=1
            )
            mock_get_client.return_value = mock_client

            # Setup mock tokenizer behavior
            mock_tokenizer = Mock()
            mock_tokenizer.apply_chat_template.return_value = torch.tensor([[1, 2, 3]])

            def mock_decode(tokens: torch.Tensor, *_args: Any, **_kwargs: Any) -> str:
                result = ""
                for token in tokens:
                    if token == 4:
                        result += _RESULT
                    else:
                        raise ValueError("Unexpected token")
                return result

            mock_tokenizer.decode = mock_decode
            mock_get_tokenizer.return_value = mock_tokenizer

            # Assert output being passed through correctly
            assert (await huggingface_local_chat.get_response(_PROMPT)) == [_RESULT]


#######################################################################################
#                                     FIXTURES                                        #
#######################################################################################


@pytest.fixture
def cached_huggingface_local(
    hf_token: str, data_path: Path
) -> Iterator[CompletionModel]:
    database = data_path / "_copy_huggingface_local.db"
    shutil.copyfile(data_path / "huggingface_local.db", database)

    yield CachedCompletionModel(
        model=HuggingfaceCompletionLocal(
            model_id=MODEL_MISTRAL_7B,
            api_key_env=hf_token,
            max_tokens=10,
            temperature=0.2,
        ),
        database=database,
    )

    os.remove(database)


@pytest.fixture
def huggingface_local_completion(hf_token: str) -> CompletionModel:
    return HuggingfaceCompletionLocal(
        api_key_env=hf_token,
        model_id=ZEPHYR_7B_MODEL,
        use_cuda=False,
    )


@pytest.fixture
def huggingface_local_chat(hf_token: str) -> ChatModel:
    return HuggingfaceChatLocal(
        api_key_env=hf_token,
        model_id=ZEPHYR_7B_MODEL,
        use_cuda=False,
    )


@pytest.fixture
def zephyr_7b_tokenizer() -> PreTrainedTokenizerBase:
    tokenizer = PreTrainedTokenizerBase()
    tokenizer.chat_template = (
        "{% for message in messages %}\n{% if message['role'] == 'user' %}\n"
        "{{ '<|user|>\n' + message['content'] + eos_token }}\n{% elif message['role'] == 'system' %}\n"
        "{{ '<|system|>\n' + message['content'] + eos_token }}\n{% elif message['role'] == 'assistant' %}\n"
        "{{ '<|assistant|>\n'  + message['content'] + eos_token }}\n{% endif %}\n"
        "{% if loop.last and add_generation_prompt %}\n{{ '<|assistant|>' }}\n{% endif %}\n{% endfor %}"
    )
    tokenizer.eos_token = "</s>"
    return tokenizer



# -----------------------------------------------------------------------------
# © 2024 Boston Consulting Group. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
Implementation of ``LogThrottlingHandler``.
"""

import logging
import time
from collections import defaultdict

__all__ = [
    "LogThrottlingHandler",
]


class LogThrottlingHandler(logging.Handler):
    """
    A logging handler that limits the number of messages during a given time interval.

    This handler can be used with Python's native :mod:`logging` module to throttle
    the maximum number of messages sent during a given time interval.
    This is useful for preventing log spamming in high-throughput applications.

    To follow the native logging flow
    (https://docs.python.org/3/howto/logging.html#logging-flow),
    functionality is implemented across the :meth:`.filter` and :meth:`.emit` methods.

    Example:

    .. code-block:: python

        import logging
        from artkit.util import LogThrottlingHandler

        console_handler = logging.StreamHandler()
        throttling_handler = LogThrottlingHandler(
            console_handler,
            interval=5,
            max_messages=5
        )
        logger = logging.getLogger()
        logger.addHandler(throttling_handler)
    """

    def __init__(
        self, handler: logging.Handler, interval: float, max_messages: int
    ) -> None:
        """
        :param handler: the handler to wrap
        :param interval: the minimum interval in seconds between log messages
          with the same message
        :param max_messages: the maximum number of messages to log within the interval
        """
        super().__init__()
        self.handler = handler
        self.interval = interval
        self.max_messages = max_messages
        self.log_counts: defaultdict[str, tuple[int, float, list[str]]] = defaultdict(
            lambda: (0, 0, [])
        )

    def filter(self, record: logging.LogRecord) -> bool:
        """
        Filter a log record based on the throttling settings.

        :param record: the log record to filter
        :return: ``True`` if the max messages are not exceeded
          within the time interval, otherwise ``False``
        """

        # if any other filter was registered and returns False, return False
        if not super().filter(record):
            return False

        count, last_log_time, buffer = self.log_counts[record.msg]

        current_time = time.time()
        if current_time - last_log_time > self.interval:
            count = 0
            last_log_time = current_time
            buffer = []

        self.log_counts[record.msg] = (count, last_log_time, buffer)

        if count > self.max_messages:
            return False
        return True

    def emit(self, record: logging.LogRecord) -> None:
        """
        Emit a record if the max messages are not exceeded within the time interval.

        :param record: the log record to emit
        """
        count, last_log_time, buffer = self.log_counts[record.msg]

        if count < self.max_messages:
            self.handler.emit(record)
            buffer.append(record.msg)
            count += 1
        elif count == self.max_messages:
            ellipsis_record = self._create_ellipsis_record(record)
            self.handler.emit(ellipsis_record)
            count += 1

        self.log_counts[record.msg] = (count, last_log_time, buffer)

    @staticmethod
    def _create_ellipsis_record(record: logging.LogRecord) -> logging.LogRecord:
        """
        Create a log record with a custom message to indicate throttling.

        :param record: the original log record
        :return: a new log record with the custom message message
        """
        return logging.LogRecord(
            name=record.name,
            level=record.levelno,
            pathname=record.pathname,
            lineno=record.lineno,
            msg="---- Logs truncated ----",
            args=(),
            exc_info=record.exc_info,
        )



"""
Tests for the Image class
"""

from __future__ import annotations

import logging

import pytest

from artkit.util import Image

log = logging.getLogger(__name__)


def test_image_constructor() -> None:
    # Create an image from a URL

    image_from_url = Image(url="https://media-publications.bcg.com/BCG_MONOGRAM.png")
    assert image_from_url.url == "https://media-publications.bcg.com/BCG_MONOGRAM.png"
    assert image_from_url.to_b64_string().startswith("iVBO")
    assert image_from_url.to_b64().startswith(b"iVBO")
    assert image_from_url.data.startswith(b"\x89PNG")

    # Create an image from binary data

    image_from_data = Image(data=image_from_url.data)
    assert image_from_data.data == image_from_url.data
    assert image_from_data.to_b64_string() == image_from_url.to_b64_string()
    assert image_from_data.to_b64() == image_from_url.to_b64()

    # Create an image from a URL and binary data
    with pytest.raises(
        ValueError, match=r"^Exactly one of arg data and arg url must be provided$"
    ):
        Image(  # type: ignore[call-overload]
            url="https://media-publications.bcg.com/BCG_MONOGRAM.png", data=b"aaa"
        )

    # Create an image without a URL or binary data
    with pytest.raises(
        ValueError, match=r"^Either arg data or arg url must be provided$"
    ):
        Image()  # type: ignore[call-overload]


def test_image(image: Image) -> None:
    assert image._repr_png_() == image.data
    assert str(image) == f"Image(\n    data={image.data!r}\n)"

    image_with_url = Image(url="abc")
    # We set the data field so that the image does not attempt to fetch the URL
    image_with_url._data = image.data

    assert str(image_with_url) == "Image(url='abc')"


def test_image_comparison(image: Image) -> None:

    image_copy = Image(data=image.data)
    assert image == image_copy
    assert hash(image) == hash(image_copy)

    # Set the URL field so that both image objects are different
    image_copy.url = "abc"

    assert image != image_copy
    assert hash(image) != hash(image_copy)



#!/usr/bin/env python3
"""
Make sphinx documentation using the makefile in the ``make`` subdirectory
"""

import os
import sys

if __name__ == "__main__":
    sys.path.insert(
        0, os.path.join(os.path.dirname(os.path.realpath(__file__)), "make")
    )

    from make_base import make

    make()



# -----------------------------------------------------------------------------
# © 2024 Boston Consulting Group. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
Implementation of the text producer.
"""

import logging
import re
from collections.abc import Iterable

from pytools.api import inheritdoc
from pytools.expression import (
    Expression,
    HasExpressionRepr,
    expression_from_init_params,
)
from pytools.text import TextTemplate

from .base import ChatModel

log = logging.getLogger(__name__)

__all__ = [
    "TextGenerator",
]


@inheritdoc(match="""[see superclass]""")
class TextGenerator(HasExpressionRepr):
    """
    Generates text using an LLM.

    Sets up an LLM with a system prompt that instructs the LLM to generate text items.

    The system prompt must:

    - include a set of required formatting keys, in the form of '{key}', that will be
      substituted with values before sending the system prompt to the LLM
    - instruct the LLM to accept an integer as the user prompt, and generate a number
      of texts indicated by the user prompt
    - instruct the LLM to separate generated text items by lines consisting of '#####'
    """

    #: The LLM system used to generate text items.
    llm: ChatModel

    def __init__(
        self, *, llm: ChatModel, system_prompt: str, formatting_keys: Iterable[str]
    ) -> None:
        """
        :param llm: the LLM system used to generate text
        :param system_prompt: the system prompt instructing the LLM to generate the
            text items
        :param formatting_keys: the names of the formatting keys used in the system
            prompt
        """
        super().__init__()
        self.llm = llm
        self._system_prompt_template = TextTemplate(
            format_string=system_prompt, required_keys=formatting_keys
        )

    @property
    def system_prompt(self) -> str:
        """
        The system prompt used to generate text items.
        """
        return self._system_prompt_template.format_string

    @property
    def formatting_keys(self) -> set[str]:
        """
        The formatting keys used in the system prompt.
        """
        return self._system_prompt_template.formatting_keys

    async def make_text_items(self, *, n: int, attributes: dict[str, str]) -> list[str]:
        """
        Use the LLM to generate the given number of text items, substituting the
        given attributes for the formatting keys in the system prompt.

        Calls method :meth:`parse_llm_response` to parse the LLM response into
        individual text items.

        :param n: the number of text items to generate
        :param attributes: the attributes to substitute for the formatting keys in
            the system prompt
        :return: the generated text items
        """

        # we make at least 10% more text items than requested, since the LLM sometimes
        # does not produce the full number of requested text items
        n_with_buffer = n * 11 // 10 + 1

        response = await self.llm.with_system_prompt(
            system_prompt=(
                self._system_prompt_template.format_with_attributes(**attributes)
            )
        ).get_response(message=str(n_with_buffer))

        # get the text items from the last response
        if len(response) > 1:
            log.warning("LLM returned more than one response; using the last one")
        text_items = self.parse_llm_response(response[-1])

        if len(text_items) < n:
            raise ValueError(
                f"Expected {n} text items but only got {len(text_items)} "
                f"in LLM response\n{response[-1]!r}"
            )

        return text_items[:n]

    @staticmethod
    def parse_llm_response(response: str) -> list[str]:
        """
        Parse an LLM response into individual text items.

        Unless overridden, splits the text using ``"#####"`` as the separator.
        Called by method :meth:`.make_text_items` with the LLM response.

        :param response: the LLM response to parse
        :return: a list of text items
        """
        return list(
            filter(
                # exclude empty descriptions, e.g., after a trailing "#####"
                None,
                map(
                    # strip whitespace
                    str.strip,
                    # split the response by the separator string
                    # be flexible with the number of '#' characters
                    # since the LLM may not use the precise number indicated
                    re.split(r"###+", response),
                ),
            )
        )

    def to_expression(self) -> Expression:
        """[see superclass]"""
        return expression_from_init_params(self)


#
# Auxiliary functions
#



# -----------------------------------------------------------------------------
# © 2024 Boston Consulting Group. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
Implementation of the Gemini llm module.
"""

import logging
from collections.abc import Iterator, Mapping
from contextlib import AsyncExitStack
from typing import Any, TypeVar

from pytools.api import MissingClassMeta, appenddoc, inheritdoc, subsdoc

from ...util import RateLimitException
from ..base import ChatModelConnector
from ..history import ChatHistory

log = logging.getLogger(__name__)

try:  # pragma: no covers
    from google.api_core.exceptions import TooManyRequests
    from google.generativeai import GenerationConfig, GenerativeModel, configure
    from google.generativeai.types import AsyncGenerateContentResponse

except ImportError:

    class GenerativeModel(  # type: ignore[no-redef]
        metaclass=MissingClassMeta, module="google.generativeai"
    ):
        """Placeholder class for missing ``GenerativeModel`` class."""

    class GenerationConfig(  # type: ignore[no-redef]
        metaclass=MissingClassMeta, module="google.generativeai"
    ):
        """Placeholder class for missing ``GenerationConfig`` class."""

    class AsyncGenerateContentResponse(  # type: ignore[no-redef]
        metaclass=MissingClassMeta, module="google.generativeai.types"
    ):
        """Placeholder class for missing ``AsyncGenerateContentResponse`` class."""

    # noinspection PyPep8Naming
    class configure(  # type: ignore[no-redef]
        # pragma: no cover
        metaclass=MissingClassMeta,
        module="google.generativeai",
    ):
        """Placeholder class for missing ``configure`` function."""


__all__ = [
    "GeminiChat",
]


#
# Type variables
#

T_GeminiLLM = TypeVar("T_GeminiLLM", bound="GeminiChat")

#
# Class declarations
#


@inheritdoc(match="""[see superclass]""")
class GeminiChat(ChatModelConnector[GenerativeModel]):
    """
    An asynchronous Gemini LLM.
    """

    #: If ``False``, disable all accessible safety categories; otherwise, enable them.
    safety: bool

    @subsdoc(
        # The pattern matches the row defining model_params, and move it to the end
        # of the docstring.
        pattern=r"(:param model_params: .*\n)((:?.|\n)*\S)(\n|\s)*",
        replacement=r"\2\1",
    )
    @appenddoc(to=ChatModelConnector.__init__)
    def __init__(
        self,
        *,
        model_id: str,
        api_key_env: str | None = None,
        initial_delay: float = 1,
        exponential_base: float = 2,
        jitter: bool = True,
        max_retries: int = 10,
        system_prompt: str | None = None,
        safety: bool = True,
        **model_params: Any,
    ) -> None:
        """
        :param safety: if ``False``, disable all accessible safety categories;
            otherwise, enable them (default: ``True``)
        """
        super().__init__(
            model_id=model_id,
            api_key_env=api_key_env,
            initial_delay=initial_delay,
            exponential_base=exponential_base,
            jitter=jitter,
            max_retries=max_retries,
            system_prompt=system_prompt,
            **model_params,
        )
        self.safety = safety

    @classmethod
    def get_default_api_key_env(cls) -> str:  # pragma: no cover
        """[see superclass]"""
        return "GEMINI_API_KEY"

    def get_model_params(self) -> Mapping[str, Any]:
        """[see superclass]"""
        return dict(**super().get_model_params(), safety=self.safety)

    def _make_client(self) -> GenerativeModel:  # pragma: no cover

        configure(api_key=self.get_api_key())
        return GenerativeModel(self.model_id, system_instruction=self.system_prompt)

    def _get_safety_settings(self) -> dict[str, str] | None:
        """
        Get safety settings for Gemini.

        :return: a list of dicts describing safety settings by harm category
        """

        if self.safety:
            return None
        else:
            return {
                "HARM_CATEGORY_HARASSMENT": "BLOCK_NONE",
                "HARM_CATEGORY_HATE_SPEECH": "BLOCK_NONE",
                "HARM_CATEGORY_SEXUALLY_EXPLICIT": "BLOCK_NONE",
                "HARM_CATEGORY_DANGEROUS_CONTENT": "BLOCK_NONE",
            }

    @staticmethod
    def _messages_to_gemini_format(
        user_message: str, *, history: ChatHistory | None = None
    ) -> Iterator[dict[str, Any]]:
        """
        Get the messages to send to the Gemini LLM, based on the given user prompt
        and chat history, and the system prompt for this LLM.

        :param user_message: the user prompt to send to the Gemini LLM
        :param history: the chat history to include in the messages (optional)
        :return: the messages object, in the format expected by the Gemini API
        """
        if history is not None:
            for message in history.messages:
                yield {"role": message.role, "parts": [message.text]}

        yield {"role": "user", "parts": [user_message]}

    async def get_response(
        self,
        message: str,
        *,
        history: ChatHistory | None = None,
        **model_params: dict[str, Any],
    ) -> list[str]:
        """[see superclass]"""
        async with AsyncExitStack():

            config = GenerationConfig(**{**self.model_params, **model_params})
            try:
                response = await self.get_client().generate_content_async(
                    list(self._messages_to_gemini_format(message, history=history)),
                    generation_config=config,
                    safety_settings=self._get_safety_settings(),
                )
            except TooManyRequests as e:
                raise RateLimitException(
                    "Rate limit exceeded. Please try again later."
                ) from e

        return list(self._text_from_response(response))

    @staticmethod
    def _text_from_response(
        response: AsyncGenerateContentResponse,
    ) -> Iterator[str]:
        """
        Get the text from the given response.

        :param response: the response to process
        :return: the text of the responses
        """
        for candidate in response.candidates:
            yield candidate.content.parts[0].text



"""
Test JSON utility functions.
"""

from __future__ import annotations

import logging

import pytest

from artkit.model.llm.base import ChatModel
from artkit.model.llm.util import parse_json_autofix

log = logging.getLogger(__name__)


@pytest.mark.asyncio
async def test_json_autofix(cached_openai: ChatModel) -> None:
    """
    Test JSON auto-fixing.
    """

    # Test valid JSON
    assert await parse_json_autofix(
        '{"a": 1, "b": 2}',
        model=cached_openai,
    ) == dict(a=1, b=2)

    # Test valid JSON with successful validator
    assert await parse_json_autofix(
        '{"a": 1, "b": 2}',
        model=cached_openai,
        validator=lambda _: None,
    ) == dict(a=1, b=2)

    # Test invalid JSON
    assert await parse_json_autofix(
        '{"a": 1, "b": 2',
        model=cached_openai,
    ) == dict(a=1, b=2)

    # Test valid JSON with failing validator
    assert await parse_json_autofix(
        '{"a": 1, "b": 2}',
        model=cached_openai,
        validator=lambda d: None if d["b"] == 3 else "b should be 3",
    ) == dict(a=1, b=3)

    # Test valid JSON with wrong syntax after attempting to fix it
    with pytest.raises(
        ValueError,
        match=(
            r"^Failed to fix malformed JSON with error 'Expecting value: "
            r"line 1 column 1 \(char 0\)':"
        ),
    ):
        await parse_json_autofix(
            "f = lambda x: x + 1",
            model=cached_openai,
        )

    # Test valid JSON with wrong syntax after attempting to fix it
    with pytest.raises(
        ValueError,
        match=(
            r"^Failed to fix JSON after validation error 'replace with a Python lambda "
            r"function that returns 1':\n1$"
        ),
    ):
        await parse_json_autofix(
            "1",
            model=cached_openai,
            validator=lambda _: "replace with a Python lambda function that returns 1",
        )

    # Test validation failure after attempting to fix it
    with pytest.raises(
        ValueError,
        match=(
            r"Failed to fix JSON after validation error \'b should be 3\':\n"
            r'{"a": 1, "b": 2}'
        ),
    ):
        await parse_json_autofix(
            '{"a": 1, "b": 2}',
            model=cached_openai,
            validator=lambda _: "b should be 3",
        )



from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from openai import RateLimitError

from artkit.model.llm.base import ChatModel
from artkit.model.llm.openai import OpenAIChat
from artkit.model.util import RateLimitException

_ = pytest.importorskip("google.generativeai")


@pytest.mark.asyncio
async def test_openai(openai_chat: OpenAIChat) -> None:
    # Mock openai Client
    with patch("artkit.model.llm.openai._openai.AsyncOpenAI") as mock_get_client:
        # Mock openai Client response
        mock_response = AsyncMock(
            return_value=AsyncMock(
                choices=[MagicMock(message=MagicMock(content="blue", role="assistant"))]
            )
        )

        # Set mock response as return value
        mock_get_client.return_value.chat.completions.create = mock_response

        # Call mocked model
        messages = await openai_chat.get_response(
            message="What color is the sky? Please answer in one word."
        )
        assert "blue" in messages[0].lower()


@pytest.mark.asyncio
async def test_openai_retry(
    openai_chat: OpenAIChat, caplog: pytest.LogCaptureFixture
) -> None:
    # Mock openai Client
    with patch("artkit.model.llm.openai._openai.AsyncOpenAI") as mock_get_client:
        # Set mock response as return value
        response = MagicMock()
        response.status_code = 429

        # Mock exception on method call
        mock_get_client.return_value.chat.completions.create.side_effect = (
            RateLimitError(
                message="Rate Limit exceeded",
                response=response,
                body=MagicMock(),
            )
        )

        with pytest.raises(RateLimitException):
            # Call mocked model
            await openai_chat.get_response(
                message="What color is the sky? Please answer in one word."
            )
        assert (
            mock_get_client.return_value.chat.completions.create.call_count
            == openai_chat.max_retries
        )
    assert (
        len(
            [
                record
                for record in caplog.records
                if record.message.startswith("Rate limit exceeded")
            ]
        )
        == openai_chat.max_retries
    )


@pytest.mark.asyncio
async def test_cached_openai(
    cached_openai: ChatModel,
) -> None:
    messages = await cached_openai.get_response(
        message="What color is the sky? Please answer in one word."
    )
    assert "blue" in messages[0].lower()



import os
import shutil
from collections.abc import Iterator
from pathlib import Path
from unittest.mock import AsyncMock, Mock, patch

import pytest
from google.api_core.exceptions import TooManyRequests

from artkit.model.llm import CachedChatModel
from artkit.model.llm.base import ChatModel
from artkit.model.llm.gemini import GeminiChat
from artkit.model.util import RateLimitException

_ = pytest.importorskip("google.generativeai")


@pytest.mark.asyncio
async def test_gemini(gemini_chat: GeminiChat) -> None:
    # Mock Gemini Client
    with patch("artkit.model.llm.gemini._gemini.GenerativeModel") as mock_get_client:
        # Mock Gemini Client response
        mock_response = AsyncMock(
            return_value=Mock(
                candidates=[Mock(content=Mock(parts=[Mock(text="blue")]))]
            )
        )

        # Set mock response as return value
        mock_get_client.return_value.generate_content_async = mock_response

        # Call mocked model
        messages = await gemini_chat.get_response(
            message="What color is the sky? Please answer in one word."
        )
        assert "blue" in messages[0].lower()


@pytest.mark.asyncio
async def test_gemini_retry(
    gemini_chat: GeminiChat, caplog: pytest.LogCaptureFixture
) -> None:
    # Mock Gemini Client
    with patch("artkit.model.llm.gemini._gemini.GenerativeModel") as mock_get_client:
        # Set mock response as return value
        mock_get_client.return_value.generate_content_async.side_effect = (
            TooManyRequests("Rate limit error")
        )

        with pytest.raises(RateLimitException):
            # Call mocked model
            await gemini_chat.get_response(
                message="What color is the sky? Please answer in one word."
            )
        assert (
            mock_get_client.return_value.generate_content_async.call_count
            == gemini_chat.max_retries
        )
    assert (
        len(
            [
                record
                for record in caplog.records
                if record.message.startswith("Rate limit exceeded")
            ]
        )
        == gemini_chat.max_retries
    )


@pytest.mark.asyncio
async def test_cached_gemini(
    cached_gemini: ChatModel,
) -> None:
    messages = await cached_gemini.get_response(
        message="What color is the sky? Please answer in one word."
    )
    assert "blue" in messages[0].lower()


@pytest.fixture
def gemini_chat() -> GeminiChat:
    api_key_env = "TEST"
    os.environ[api_key_env] = "test"

    return GeminiChat(
        model_id="models/gemini-pro",
        api_key_env=api_key_env,
        max_output_tokens=10,
        max_retries=2,
        initial_delay=0.1,
        exponential_base=1.5,
    )


@pytest.fixture
def cached_gemini(data_path: Path, gemini_chat: GeminiChat) -> Iterator[ChatModel]:
    database = data_path / "_copy_gemini.db"
    shutil.copyfile(data_path / "gemini.db", database)

    yield CachedChatModel(model=gemini_chat, database=database)

    os.remove(database)



# -----------------------------------------------------------------------------
# © 2024 Boston Consulting Group. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
Base implementation of vision model
"""

from abc import ABCMeta, abstractmethod
from typing import Any, Generic, TypeVar

from ....util import Image
from ...base import ConnectorMixin, GenAIModel
from ...util import retry_with_exponential_backoff

__all__ = [
    "VisionModel",
    "VisionModelConnector",
]


#
# Type variables
#
T_Client = TypeVar("T_Client")


#
# Classes
#


class VisionModel(GenAIModel, metaclass=ABCMeta):
    """
    An abstract vision model.

    Vision models are used to generate text descriptions of images.
    """

    @abstractmethod
    async def image_to_text(
        self,
        image: Image,
        prompt: str | None = None,
        **model_params: Any,
    ) -> list[str]:
        """
        Generates a text response from an image and prompt pair

        :param image: an Image object to pass to the model
        :param prompt: the text prompt to pass to the model
        :param model_params: additional optional parameters for image-to-text generation
        :return: model text response for the input image and prompt pair
        """


class VisionModelConnector(
    VisionModel, ConnectorMixin[T_Client], Generic[T_Client], metaclass=ABCMeta
):
    """
    A vision model that connects to a client.
    """

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        # Apply the retry strategy to the image_to_text method
        cls.image_to_text = (  # type: ignore[method-assign]
            retry_with_exponential_backoff(cls.image_to_text)  # type: ignore
        )



import os
from collections.abc import Generator
from typing import Any
from unittest.mock import AsyncMock, Mock, patch

import boto3
import pytest
from aiohttp import ClientResponseError
from moto import mock_aws

from artkit.model.diffusion.bedrock import TitanBedrockDiffusionModel
from artkit.model.util import RateLimitException

#######################################################################################
#                                     Constants                                       #
#######################################################################################
MODEL_ID = "amazon.titan-image-generator-v1"
REGION = "us-east-1"

MESSAGE = "A cat with wings"


@pytest.mark.asyncio
async def test_text_to_image(
    mock_bedrock_diffusion_chat: TitanBedrockDiffusionModel,
) -> None:
    with patch(
        "artkit.model.diffusion.bedrock.base._base.ClientSession.__aenter__"
    ) as MockClientSession:
        mock_post = Mock()
        mock_post.json = AsyncMock(return_value={"images": ["dGVzdA=="]})
        mock_post.text = AsyncMock()
        mock_post.return_value.status_code = 200

        mock_connection = AsyncMock()
        mock_connection.post.return_value = mock_post
        MockClientSession.return_value = mock_connection

        response = await mock_bedrock_diffusion_chat.text_to_image(text=MESSAGE)
        assert response[0].data.decode("ascii") == "test"


@pytest.mark.asyncio
async def test_rate_limit_error(
    mock_bedrock_diffusion_chat: TitanBedrockDiffusionModel,
) -> None:
    with patch(
        "artkit.model.diffusion.bedrock.base._base.ClientSession.__aenter__"
    ) as MockClientSession:
        # Set up the mock connection object
        mock_connection = AsyncMock()

        def raise_rate_limit_error() -> None:
            err = ClientResponseError(
                request_info=AsyncMock(),
                history=AsyncMock(),
                status=429,
                message="Rate limit exceeded",
            )
            raise err

        mock_connection.post.return_value.raise_for_status = raise_rate_limit_error
        MockClientSession.return_value = mock_connection

        with pytest.raises(RateLimitException):
            await mock_bedrock_diffusion_chat.text_to_image(text=MESSAGE)


#######################################################################################
#                                     FIXTURES                                        #
#######################################################################################
@pytest.fixture(scope="function")
def mock_bedrock_diffusion_chat(
    aws_credentials: Any,
) -> Generator[TitanBedrockDiffusionModel, None, None]:
    with mock_aws():
        # Mock the AWS credentials
        boto3.client("sts").get_caller_identity()

        yield TitanBedrockDiffusionModel(
            model_id=MODEL_ID,
            region=REGION,
            max_retries=2,
            initial_delay=0.1,
            exponential_base=1.5,
        )


@pytest.fixture(scope="function")
def aws_credentials() -> None:
    """Mocked AWS Credentials for moto."""
    os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"
    os.environ["AWS_SECURITY_TOKEN"] = "testing"
    os.environ["AWS_SESSION_TOKEN"] = "testing"
    os.environ["AWS_DEFAULT_REGION"] = "us-east-1"



# -----------------------------------------------------------------------------
# © 2024 Boston Consulting Group. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
Implementation of llm module.
"""

import copy
import logging
from abc import ABCMeta, abstractmethod
from typing import Any, Generic, TypeVar, final

from typing_extensions import Self

from pytools.api import appenddoc, inheritdoc, subsdoc

from ...base import ConnectorMixin, GenAIModel
from ...util import retry_with_exponential_backoff
from ..history import ChatHistory

log = logging.getLogger(__name__)


__all__ = [
    "ChatModel",
    "ChatModelConnector",
    "CompletionModel",
    "CompletionModelConnector",
]

#
# Type variables
#

T_ChatModel = TypeVar("T_ChatModel", bound="ChatModel")
T_Client = TypeVar("T_Client")

#
# Class declarations
#


class CompletionModel(GenAIModel, metaclass=ABCMeta):
    """
    A Large Language Model (LLM) that generates text completions.
    """

    @abstractmethod
    async def get_completion(
        self, *, prompt: str, **model_params: dict[str, Any]
    ) -> str:
        """
        Generate a text completion for the given prompt.

        :param prompt: the prompt to pass to the completion model
        :param model_params: additional parameters for the completion model
        :return: the text generated by the completion model, excluding the prompt
        """


class CompletionModelConnector(
    CompletionModel, ConnectorMixin[T_Client], Generic[T_Client], metaclass=ABCMeta
):
    """
    A connector to a Large Language Model (LLM) that generates text completions.
    """

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        # Apply the retry strategy to the image_to_text method
        cls.get_completion = (  # type: ignore[method-assign]
            retry_with_exponential_backoff(cls.get_completion)  # type: ignore
        )


class ChatModel(GenAIModel, metaclass=ABCMeta):
    """
    A Large Language Model (LLM) that generates responses for text prompts.

    Uses an optional system prompt to set up the model.
    """

    @property
    @abstractmethod
    def system_prompt(self) -> str | None:
        """
        The system prompt used to set up the LLM system.
        """

    @abstractmethod
    def with_system_prompt(self, system_prompt: str) -> Self:
        """
        Set the system prompt for the LLM system.

        :param system_prompt: the system prompt to use
        :return: a new LLM system with the system prompt set
        """

    @abstractmethod
    async def get_response(
        self,
        message: str,
        *,
        history: ChatHistory | None = None,
        **model_params: dict[str, Any],
    ) -> list[str]:
        """
        Get a response, or multiple alternative responses, from the chat system.

        :param message: the user prompt to pass to the chat system
        :param history: the chat history preceding the message
        :param model_params: additional parameters for the chat system
        :return: the response or alternative responses generated by the chat system
        :raises RequestLimitException: if an error occurs while communicating
            with the chat system
        """


@inheritdoc(match="""[see superclass]""")
class ChatModelConnector(
    ChatModel, ConnectorMixin[T_Client], Generic[T_Client], metaclass=ABCMeta
):
    """
    A chat system that connects to a client.
    """

    #: The system prompt used to set up the chat system.
    _system_prompt: str | None

    @subsdoc(
        # The pattern matches the row defining model_params, and move it to the end
        # of the docstring.
        pattern=r"(:param model_params: .*\n)((:?.|\n)*\S)(\n|\s)*",
        replacement=r"\2\1",
    )
    @appenddoc(to=ConnectorMixin.__init__)
    def __init__(
        self,
        *,
        model_id: str,
        api_key_env: str | None = None,
        initial_delay: float | None = None,
        exponential_base: float | None = None,
        jitter: bool | None = None,
        max_retries: int | None = None,
        system_prompt: str | None = None,
        **model_params: Any,
    ) -> None:
        """
        :param system_prompt: the system prompt to initialise the chat system with
          (optional)
        """
        super().__init__(
            model_id=model_id,
            api_key_env=api_key_env,
            initial_delay=initial_delay,
            exponential_base=exponential_base,
            jitter=jitter,
            max_retries=max_retries,
            **model_params,
        )
        self._system_prompt = system_prompt

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        # Apply the retry strategy to the image_to_text method
        cls.get_response = (  # type: ignore[method-assign]
            retry_with_exponential_backoff(cls.get_response)  # type: ignore
        )

    @property
    @final
    def system_prompt(self) -> str | None:
        """[see superclass]"""
        return self._system_prompt

    def with_system_prompt(self, system_prompt: str) -> Self:
        """
        Set the system prompt for the LLM system.

        :param system_prompt: the system prompt to use
        :return: a new LLM system with the system prompt set
        """
        llm_copy = copy.copy(self)
        llm_copy._system_prompt = system_prompt
        return llm_copy



"""
General tests for the 'model' module.
"""

from __future__ import annotations

import logging
import os
from typing import Any

import pytest

from artkit.model.base import ConnectorMixin

log = logging.getLogger(__name__)


def test_api_key() -> None:

    # A dummy model
    class MyModel(ConnectorMixin[Any]):

        @classmethod
        def get_default_api_key_env(cls) -> str:
            return "MY_KEY"

        def _make_client(self) -> Any:
            pass

    # Create a dummy API key in the environment
    os.environ["MY_CUSTOM_KEY"] = "my_custom_key"

    # Create the model with a custom API key environment variable
    model = MyModel(model_id="my_model", api_key_env="MY_CUSTOM_KEY")

    # Test the API key
    assert model.get_api_key() == "my_custom_key"

    # Create a dummy API key in the environment
    os.environ["MY_KEY"] = "my_key"

    # Create the model with the default API key environment variable
    model = MyModel(model_id="my_model")

    # Test the API key
    assert model.get_api_key() == "my_key"

    # Removing the dummy API key leads to an error when trying to get the API key
    del os.environ["MY_KEY"]
    with pytest.raises(
        ValueError,
        match=(
            r"^The environment variable MY_KEY for the API key of model 'my_model' is "
            r"not set\. Please set the environment variable to your API key, or revise "
            r"arg api_key_env to the correct environment variable name\.$"
        ),
    ):
        model.get_api_key()



import os
from collections.abc import Generator
from typing import Any
from unittest.mock import AsyncMock, Mock, patch

import boto3
import pytest
from httpx import HTTPStatusError, Response
from moto import mock_aws

from artkit.model.llm.bedrock import TitanBedrockChat
from artkit.model.util import RateLimitException

#######################################################################################
#                                     Constants                                       #
#######################################################################################
MODEL_ID = "amazon.titan-text-lite-v1"
REGION = "us-east-1"

MESSAGE = "What is the color of the sky?"
RESPONSE_TEXT = "The sky is blue."
PROMPT_COMPLETION = "blue"


@pytest.mark.asyncio
async def test_get_response(mock_bedrock_chat: TitanBedrockChat) -> None:
    with patch(
        "artkit.model.llm.base.HTTPXChatConnector.get_client"
    ) as MockClientSession:
        mock_request = Mock()
        mock_request.json = Mock(
            return_value={"results": [{"outputText": RESPONSE_TEXT}]}
        )
        mock_request.text = AsyncMock()
        mock_request.return_value.status_code = 200

        mock_connection = AsyncMock()
        mock_connection.request.return_value = mock_request
        MockClientSession.return_value = mock_connection

        response = await mock_bedrock_chat.get_response(message=MESSAGE)
        assert response == [RESPONSE_TEXT]


@pytest.mark.asyncio
async def test_rate_limit_error(mock_bedrock_chat: TitanBedrockChat) -> None:
    with patch(
        "artkit.model.llm.base.HTTPXChatConnector.get_client"
    ) as MockClientSession:
        # Set up the mock connection object
        mock_connection = AsyncMock()

        def raise_rate_limit_error() -> None:
            err = HTTPStatusError(
                request=Mock(),
                response=Mock(spec=Response, status_code=429),
                message="Rate limit exceeded",
            )
            raise err

        mock_connection.request.return_value.raise_for_status = raise_rate_limit_error
        MockClientSession.return_value = mock_connection

        with pytest.raises(RateLimitException):
            await mock_bedrock_chat.get_response(message=MESSAGE)


@pytest.mark.asyncio
async def test_invalid_request_error(mock_bedrock_chat: TitanBedrockChat) -> None:
    with patch(
        "artkit.model.llm.base.HTTPXChatConnector.get_client"
    ) as MockClientSession:
        # Set up the mock connection object
        mock_connection = AsyncMock()

        def raise_invalid_request_error() -> None:
            err = HTTPStatusError(
                request=Mock(),
                response=Mock(spec=Response, status_code=422),
                message="Invalid request",
            )
            raise err

        mock_connection.request.return_value.raise_for_status = (
            raise_invalid_request_error
        )
        MockClientSession.return_value = mock_connection

        with pytest.raises(ValueError):
            await mock_bedrock_chat.get_response(message=MESSAGE)


@pytest.mark.asyncio
async def test_unexpected_error(mock_bedrock_chat: TitanBedrockChat) -> None:
    with patch(
        "artkit.model.llm.base.HTTPXChatConnector.get_client"
    ) as MockClientSession:
        # Set up the mock connection object
        mock_connection = AsyncMock()

        def raise_unexpected_error() -> None:
            err = HTTPStatusError(
                request=Mock(),
                response=Mock(spec=Response, status_code=500),
                message="Internal server error",
            )
            raise err

        mock_connection.request.return_value.raise_for_status = raise_unexpected_error
        MockClientSession.return_value = mock_connection

        with pytest.raises(HTTPStatusError):
            await mock_bedrock_chat.get_response(message=MESSAGE)


@pytest.mark.asyncio
async def test_response_parsing(mock_bedrock_chat: TitanBedrockChat) -> None:
    response = Mock(spec=Response)
    response_body = {"results": [{"outputText": RESPONSE_TEXT}]}
    response.json = Mock(return_value=response_body)
    responses = mock_bedrock_chat.parse_httpx_response(response=response)
    assert responses == [RESPONSE_TEXT]


#######################################################################################
#                                     FIXTURES                                        #
#######################################################################################
@pytest.fixture(scope="function")
def mock_bedrock_chat(aws_credentials: Any) -> Generator[TitanBedrockChat, None, None]:
    with mock_aws():
        # Mock the AWS credentials
        boto3.client("sts").get_caller_identity()

        yield TitanBedrockChat(
            model_id=MODEL_ID,
            region=REGION,
            initial_delay=0.1,
            exponential_base=1.5,
            max_retries=2,
        )


@pytest.fixture(scope="function")
def aws_credentials() -> None:
    """Mocked AWS Credentials for moto."""
    os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"
    os.environ["AWS_SECURITY_TOKEN"] = "testing"
    os.environ["AWS_SESSION_TOKEN"] = "testing"
    os.environ["AWS_DEFAULT_REGION"] = "us-east-1"



# -----------------------------------------------------------------------------
# © 2024 Boston Consulting Group. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
Huggingface LLM systems.
"""

from __future__ import annotations

import logging
from typing import Any, TypeVar, cast

from pytools.api import MissingClassMeta, inheritdoc

from ...util import RateLimitException
from ..base import HTTPXChatConnector
from ..history import ChatHistory, ChatMessage, SystemPrompt, UserMessage
from .base import (
    HuggingfaceChatConnectorMixin,
    HuggingfaceCompletionConnectorMixin,
    HuggingfaceRemoteConnectorMixin,
)

try:  # pragma: no cover
    from huggingface_hub import AsyncInferenceClient
except ImportError:

    class AsyncInferenceClient(  # type: ignore
        metaclass=MissingClassMeta, module="huggingface_hub"
    ):
        """Placeholder class for missing ``AsyncInferenceClient`` class."""


try:
    from aiohttp import ClientResponseError
except ImportError:

    class ClientResponseError(  # type: ignore
        metaclass=MissingClassMeta, module="aiohttp"
    ):
        """Placeholder class for missing ``ClientResponseError`` class."""


try:
    from httpx import Response

except ImportError:

    class Response(metaclass=MissingClassMeta, module="httpx"):  # type: ignore
        """Placeholder class for missing ``Response`` class."""


log = logging.getLogger(__name__)

__all__ = [
    "HuggingfaceCompletion",
    "HuggingfaceChat",
    "HuggingfaceURLChat",
]


#
# Type variables
#

T_Client = TypeVar("T_Client")

#
# Class declarations
#


@inheritdoc(match="""[see superclass]""")
class HuggingfaceCompletion(
    HuggingfaceRemoteConnectorMixin,
    HuggingfaceCompletionConnectorMixin[AsyncInferenceClient],
):
    """
    Asynchronous Huggingface remote API.
    """

    async def _make_completion(
        self, prompt: str, **model_params: dict[str, Any]
    ) -> str:  # pragma: no cover
        """[see superclass]"""

        try:
            return cast(
                str,
                await self.get_client().text_generation(
                    prompt,
                    **{**self.get_model_params(), **model_params},
                ),
            )
        except ClientResponseError as e:
            if e.status == 429:
                raise RateLimitException(
                    "Rate limit exceeded. Please try again later."
                ) from e
            raise e


@inheritdoc(match="""[see superclass]""")
class HuggingfaceChat(
    HuggingfaceRemoteConnectorMixin,
    HuggingfaceChatConnectorMixin[AsyncInferenceClient],
):
    """
    Asynchronous Huggingface API LLM.

    Examples:

    Model usage where chat template is implicitly derived from the tokenizer:

    .. code-block:: python

        import artkit.api as ak

        chat = ak.HuggingfaceChat(
            model_id="meta-llama/Meta-Llama-3-8B-Instruct",
            api_key_env="HF_TOKEN",
            template_kwargs={"add_generation_prompt": True}
        )

        response = await chat.get_response("Hello, how are you?")

    Set your own chat template:

    .. code-block:: python

        chat = ak.HuggingfaceChat(
            model_id="meta-llama/Meta-Llama-3-8B-Instruct",
            api_key_env="HF_TOKEN",
            custom_chat_template=(
                "{% set loop_messages = messages %}"
                "{% for message in loop_messages %}"
                "{% set content = '' + message['role'] + '\\n\\n' + message['content'] | trim + '' %}"
                "{% if loop.index0 == 0 %}{% set content = bos_token + content %}{% endif %}"
                "{{ content }}"
                "{% endfor %}"
                "{% if add_generation_prompt %}"
                "{{ 'assistant\\n\\n' }}"
                "{% endif %}"
            ),
            template_kwargs={"add_generation_prompt": True}
        )

        response = await chat.get_response("Hello, how are you?")
    """  # noqa: W505

    async def _get_client_response(
        self, messages: list[dict[str, str]], **model_params: dict[str, Any]
    ) -> list[str]:
        """[see superclass]"""

        messages = self.get_tokenizer().apply_chat_template(
            messages,
            tokenize=False,
            **self.chat_template_params,
        )
        try:
            response = await self.get_client().text_generation(
                messages,
                # return_full_text=False, stream=False, details=False,
                # ensure that the prompt is not included in the response
                # and that the response is a string
                **{
                    **{"return_full_text": False, "stream": False, "details": False},
                    # Can't use self.model_params() here since additional
                    # unrelated params are incorporated there
                    **self.model_params,
                    **model_params,
                },
            )
        except ClientResponseError as e:
            if e.status == 429:
                raise RateLimitException(
                    "Rate limit exceeded. Please try again later."
                ) from e
            raise
        if not isinstance(response, str):
            """
            stream=False, details=False, means the response should be a string
            we do not support other response types as of now as we currently
            only handle string responses for more information see:
            https://huggingface.co/docs/huggingface_hub/main/en/package_reference/ \
                inference_client#huggingface_hub.InferenceClient.text_generation
            """
            raise ValueError(f"Unexpected response type: {type(response)}")

        return [response.strip()]


@inheritdoc(match="""[see superclass]""")
class HuggingfaceURLChat(
    HTTPXChatConnector,
):
    """
    Asynchronous Huggingface API LLM using an URL as model_id.
    This class is used to interact with dedicated huggingface
    inference endpoints that use a URL as a model id.
    For regular hugging face models, use :class:`HuggingfaceChat`.

    .. note::

        This class uses the OpenAI API reference in the background
        to interact with the Huggingface API. Therefore, parameter
        names and behavior are defined here
        https://platform.openai.com/docs/api-reference/chat/create.
        This is done to avoid having to use a local tokenizer,
        while also wanting to pass a chat history.

    Examples:

    .. code-block:: python

        import artkit.api as ak

        chat = ak.HuggingfaceChat(
            model_id="http:/huggingface.com/v1/chat/completions",
        )

        response = await chat.get_response("Hello, how are you?")
    """

    @classmethod
    def get_default_api_key_env(cls) -> str:
        """[see superclass]"""
        return "HF_TOKEN"

    def build_request_arguments(
        self,
        message: str,
        *,
        history: ChatHistory | None = None,
        **model_params: dict[str, Any],
    ) -> dict[str, Any]:
        """[see superclass]"""

        messages: list[ChatMessage] = []
        # Handle system prompt
        if self.system_prompt:
            messages = [SystemPrompt(self.system_prompt)]

        # Handle chat history
        if history:
            messages.extend(history.messages)

        # Handle current message
        messages.append(UserMessage(message))

        # convert messages to the format expected by the API
        converted_messages = [
            {"role": message.role, "content": message.text} for message in messages
        ]

        api_suffix = "/v1/chat/completions"
        url = (
            self.model_id
            if self.model_id.endswith(api_suffix)
            else (self.model_id + api_suffix)
        )
        return dict(
            method="POST",  # httpx provides no enum for this, consider adding our own
            url=url,
            headers={
                "Authorization": f"Bearer {self.get_api_key()}",
                "Content-Type": "application/json",
            },
            json={
                "model": "tgi",
                "messages": converted_messages,
                **{**self.model_params, **model_params},
            },
        )

    def parse_httpx_response(
        self,
        response: Response,
    ) -> list[str]:
        """[see superclass]"""
        response_body = response.json()
        result = []
        for choice in response_body["choices"]:
            message = choice["message"]
            if message["role"] != "assistant":
                log.warning(
                    "Expected only assistant messages, but got completion choice "
                    f"{choice!r}"
                )
            result.append(str(message["content"]))
        return result



# -----------------------------------------------------------------------------
# © 2024 Boston Consulting Group. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
Bedrock Diffusion systems.
"""
from __future__ import annotations

import logging
from abc import ABCMeta, abstractmethod
from contextlib import AsyncExitStack
from typing import Any, TypeVar

from pytools.api import MissingClassMeta, appenddoc, inheritdoc, subsdoc

from .....util import Image
from ....util import RateLimitException
from ...base import DiffusionModelConnector

log = logging.getLogger(__name__)

try:
    from aiohttp import ClientResponseError, ClientSession
    from botocore.auth import SigV4Auth
    from botocore.awsrequest import AWSRequest
    from botocore.session import Session

except ImportError:

    class ClientResponseError(metaclass=MissingClassMeta, module="aiohttp"):  # type: ignore
        """Placeholder class for missing ``ClientResponseError`` class."""

    class ClientSession(metaclass=MissingClassMeta, module="aiohttp"):  # type: ignore
        """Placeholder class for missing ``ClientSession`` class."""

    class SigV4Auth(metaclass=MissingClassMeta, module="SigV4Auth"):  # type: ignore
        """Placeholder class for missing ``SigV4Auth`` class."""

    class AWSRequest(metaclass=MissingClassMeta, module="AWSRequest"):  # type: ignore
        """Placeholder class for missing ``AWSRequest`` class."""

    class Session(metaclass=MissingClassMeta, module="Session"):  # type: ignore
        """Placeholder class for missing ``Session`` class."""


__all__ = ["BaseBedrockDiffusionModel"]

#
# Type variables
#

T_BaseBedrockDiffusionModel = TypeVar(
    "T_BaseBedrockDiffusionModel", bound="BaseBedrockDiffusionModel"
)

#
# Class declarations
#
log = logging.getLogger(__name__)


@inheritdoc(match="""[see superclass]""")
class BaseBedrockDiffusionModel(DiffusionModelConnector[None], metaclass=ABCMeta):
    """
    Base class for Bedrock LLMs.
    """

    region: str | None

    @classmethod
    def get_default_api_key_env(cls) -> str:
        """[see superclass]"""
        return ""

    @classmethod
    def _make_client(self) -> None:
        """[see superclass]"""
        return None

    @subsdoc(
        # The pattern matches the row defining model_params, and move it to the end
        # of the docstring.
        pattern=r"(:param model_params: .*\n)((:?.|\n)*\S)(\n|\s)*",
        replacement=r"\2\1",
    )
    @appenddoc(to=DiffusionModelConnector.__init__)
    def __init__(
        self,
        *,
        model_id: str,
        api_key_env: str | None = None,
        initial_delay: float | None = None,
        exponential_base: float | None = None,
        jitter: bool | None = None,
        max_retries: int | None = None,
        region: str,
        **model_params: Any,
    ) -> None:
        """
        Initialize the BaseBedrockDiffusionModel.

        :param region: The AWS region.
        :raises CredentialsNotFoundError: if unable to find AWS Credentials.
        """
        super().__init__(
            model_id=model_id,
            api_key_env=api_key_env,
            initial_delay=initial_delay,
            exponential_base=exponential_base,
            jitter=jitter,
            max_retries=max_retries,
            model_params=model_params,
        )
        self.region = region if region else "us-east-1"
        self.session = Session()
        self.credentials = self.session.get_credentials()
        self.auth = SigV4Auth(self.credentials, "bedrock", self.region)
        self.endpoint = f"https://bedrock-runtime.{self.region}.amazonaws.com/model/{self.model_id}/invoke"

    async def get_response_from_model(self, body: str) -> dict[str, Any]:
        """
        Gets a response from the model

        :param body: The body of the request.
        :return: The response from the model.
        """
        request = AWSRequest(
            method="POST",
            url=self.endpoint,
            data=body,
            headers={"content-type": "application/json"},
        )
        self.auth.add_auth(request)
        prepped_request = request.prepare()
        headers = dict(prepped_request.headers.items())
        async with AsyncExitStack():
            async with ClientSession(headers=headers) as aio_session:
                response = await aio_session.post(
                    url=prepped_request.url, data=prepped_request.body
                )
                response_text = await response.text()
                try:
                    # Raises exception if response status is not 200
                    response.raise_for_status()
                except ClientResponseError as e:
                    if e.status == 429:
                        raise RateLimitException(
                            "Rate limit exceeded. Please try again later."
                        ) from e
                    elif e.status == 422:
                        raise ValueError(
                            f"Invalid request. Please check the request parameters. {response_text}"
                        ) from e
                    raise

        response_body: dict[str, Any] = await response.json()
        return response_body

    @abstractmethod
    async def text_to_image(
        self, text: str, **model_params: dict[str, Any]
    ) -> list[Image]:
        """
        Convert JSON text payload to images using the diffusion model.

        :param text: The payload for the specific diffusion model.
        :param model_params: Additional model parameters.
        :return: List of Image that corresponds to the output in the model response.
        """
        pass



# -----------------------------------------------------------------------------
# © 2024 Boston Consulting Group. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
Convenience module to import all classes and functions using single package alias.

Recommended usage is to import this module as ``ak``.

Example:

.. code-block:: python

    import artkit.api as ak

    ak.run(
        ak.step("add", lambda x: dict(x=x + 1)),
        input=[dict(x=1), dict(x=10)],
    )

"""

__globals_original = set(globals().keys())


from .flow import *
from .model.diffusion import *
from .model.diffusion.base import DiffusionModel
from .model.diffusion.bedrock import *
from .model.diffusion.openai import *
from .model.llm import *
from .model.llm.anthropic import *
from .model.llm.base import ChatModel, CompletionModel
from .model.llm.bedrock import *
from .model.llm.gemini import *
from .model.llm.groq import *
from .model.llm.huggingface import *
from .model.llm.multi_turn import *
from .model.llm.openai import *
from .model.llm.util import *
from .model.llm.vertexai import *
from .model.vision import *
from .model.vision.base import VisionModel
from .model.vision.openai import *
from .util import *


def __adjust_namespace_for_sphinx() -> None:  # pragma: no cover
    import os
    from collections import defaultdict

    from pytools.api import public_module_prefix

    # Remove symbols that are not functions or classes
    for name, obj in list(globals().items()):
        if not callable(obj) and not name.startswith("__"):
            del globals()[name]

    # Check if we are building the Sphinx documentation
    if "SPHINX_BUILD" not in os.environ:
        # we are not building the documentation - no adjustments needed
        return

    # We are building the Sphinx documentation

    # Get the namespace of this module
    globals_ = globals()

    # Identify all imported symbols
    imported_symbols = set(globals_.keys()) - __globals_original  # noqa: F405

    # Remove our internal symbols from items to be documented
    imported_symbols.remove("__globals_original")
    imported_symbols.remove(__adjust_namespace_for_sphinx.__name__)  # noqa: F405

    # Add all imported symbols to the module docstring
    imported_objects = [globals_[symbol] for symbol in imported_symbols]

    # Group the imported objects by their module and an optional module alias
    imported_objects_by_module: dict[tuple[str, str | None], list[str]] = defaultdict(
        list
    )
    for obj in imported_objects:
        imported_objects_by_module[(public_module_prefix(obj.__module__), None)].append(
            obj.__qualname__
        )

    # Special case: mirror the fluxus module
    import fluxus.functional as fx

    imported_objects_by_module[("fluxus.functional", "artkit.flow")] = sorted(
        (
            name
            for name, obj in vars(fx).items()
            if not name.startswith("_") and callable(obj)
        ),
        key=lambda s: s.lower(),
    )

    # Create the documentation string, with separate sections for each module
    doc = [
        f"\n* :mod:`{alias or module}`   \n\n"
        + "\n".join(f"  - :obj:`~{module}.{name}`" for name in names)
        for (module, alias), names in sorted(
            # Sort by module then by name
            imported_objects_by_module.items(),
            key=lambda x: (
                # sort by alias, if it exists, otherwise by name
                x[0][1] or x[0][0],
                x[1],
            ),
        )
    ]
    globals_["__doc__"] += (
        "Below is a list of all symbols exposed through ``artkit.api``:\n\n"
        + "\n".join(doc)
        + "\n"
    )

    # remove all imported and documented symbols from the global namespace
    for symbol in imported_symbols:
        del globals_[symbol]


# Adjust the namespace for Sphinx documentation builds
__adjust_namespace_for_sphinx()

# Clean up the namespace
del __adjust_namespace_for_sphinx, __globals_original



"""
Validate the Python code in the module.
"""

from pytools.api import validate__all__declarations


def test__all__declarations() -> None:
    import artkit
    import artkit.flow as flow

    # artkit.flow exports 3rd party symbols from the fluxus package, so we need
    # to set the modules to artkit.flow to make this test succeed

    flow_namespace = vars(flow)
    original_modules = {
        name: obj.__module__
        for name, obj in flow_namespace.items()
        if not name.startswith("_")
    }
    try:
        # Set the module of the symbols to artkit.flow
        for name in original_modules:
            flow_namespace[name].__module__ = flow.__name__

        validate__all__declarations(artkit)
    finally:
        # Restore the original modules
        for name, module in original_modules.items():
            flow_namespace[name].__module__ = module



"""
Cache tests.
"""

from __future__ import annotations

import logging
import time
from datetime import datetime, timezone

import pytest

from artkit.model.cache import CacheDB

log = logging.getLogger(__name__)


@pytest.fixture(scope="function")
def cache() -> CacheDB:
    """
    Fixture for the CacheDB class.
    """
    return CacheDB(database=":memory:")


def test_cache_db(cache: CacheDB) -> None:
    """
    Test the CacheDB class.
    """

    # Constants: model identifiers
    model_a = "model a"
    model_b = "model b"

    # Constants: prompts
    prompt_a = "prompt a"
    prompt_b = "prompt b"

    # Constants: responses
    response_single = "test_response"
    response_multi = ["test_response 1", "test_response 2"]

    # Constants: model parameters
    params_a: dict[str, str | int | float | bool] = dict(
        a_int=2, a_float=2.0, a_str="2", a_bool=True
    )
    params_b: dict[str, str | int | float | bool] = dict(
        b_int=3, b_float=3.0, b_str="3", b_bool=False
    )
    params_ab = {**params_a, **params_b}

    # Test the cache database
    assert cache.conn is not None

    # Try to get a non-existent model response
    assert cache.get_entry(model_id=model_a, prompt=prompt_a) is None

    # Add a new single response
    cache.add_entry(
        model_id=model_a,
        prompt=prompt_a,
        responses=response_single,
    )

    # Get the single response
    assert cache.get_entry(model_id=model_a, prompt=prompt_a) == [response_single]

    # Fail if we add model parameters that are not associated with the cache entry
    assert cache.get_entry(model_id=model_a, prompt=prompt_a, **params_a) is None

    # Fail if we use a different model identifier
    assert cache.get_entry(model_id=model_b, prompt=prompt_a) is None

    # Add a multi response with model parameters
    cache.add_entry(
        model_id=model_a, prompt=prompt_b, responses=response_multi, **params_a
    )

    # Get the multi response with model parameters
    assert (
        cache.get_entry(model_id=model_a, prompt=prompt_b, **params_a) == response_multi
    )

    # Fail if we use model parameters that are not associated with the cache entry
    assert cache.get_entry(model_id=model_a, prompt=prompt_b, **params_b) is None

    # Fail if we use additional model parameters that are not associated with the cache
    # entry
    assert cache.get_entry(model_id=model_a, prompt=prompt_b, **params_ab) is None

    # Fail if we do not add model parameters that are associated with the cache entry
    assert cache.get_entry(model_id=model_a, prompt=prompt_b) is None


def test_cache_clearing_before(cache: CacheDB) -> None:
    """
    Test clearing the cache before a given datetime.
    """

    cache.clear()

    # Constants: model identifiers
    model_a = "model a"
    model_b = "model b"

    # We create 4 cache entries
    n_entries = 4

    # Constants: prompts
    prompts = [f"prompt {i}" for i in range(n_entries)]

    # Constants: responses
    responses = [[f"response {i}"] for i in range(n_entries)]

    # Constants: model parameters
    params: list[dict[str, str | int | float | bool]] = [
        dict(a_int=i, a_text=f"parameter {i}", a_float=i + 0.5, a_text_shared="shared")
        for i in range(n_entries)
    ]

    # Create a new cache database
    cache.add_entry(
        model_id=model_a, prompt=prompts[0], responses=responses[0], **params[0]
    )
    time.sleep(1)
    timestamp_0 = now()
    time.sleep(1)

    # Create additional cache entries
    for i in range(1, n_entries):
        cache.add_entry(
            model_id=model_a, prompt=prompts[i], responses=responses[i], **params[i]
        )

    # Check that all entries are in the cache
    for i in range(n_entries):
        print(f"Checking entry {i}")
        assert (
            cache.get_entry(
                model_id=model_a,
                prompt=prompts[i],
                **params[i],
            )
            == responses[i]
        )

    # Clear the entry with the oldest creation time
    cache.clear(created_before=timestamp_0)

    # Check that entry 0 is not in the cache, but the others are
    assert cache.get_entry(model_id=model_a, prompt=prompts[0], **params[0]) is None
    for i in range(1, n_entries):
        assert (
            cache.get_entry(model_id=model_a, prompt=prompts[i], **params[i])
            == responses[i]
        )

    time.sleep(1)
    timestamp_1 = now()

    # Access entry 1
    assert (
        cache.get_entry(model_id=model_a, prompt=prompts[1], **params[1])
        == responses[1]
    )

    # Clear all entries accessed before timestamp_1
    cache.clear(accessed_before=timestamp_1, model_id=model_a)

    # Confirm only entry 1 is in the cache
    for i in range(n_entries):
        entry = cache.get_entry(model_id=model_a, prompt=prompts[i], **params[i])
        if i == 1:
            assert entry == responses[i]
        else:
            assert entry is None

    # Clear the cache
    cache.clear(model_id=model_b)

    # Confirm that we still have entry 1 in the cache
    assert (
        cache.get_entry(model_id=model_a, prompt=prompts[1], **params[1])
        == responses[1]
    )
    assert cache.count_entries() == {model_a: 1}

    # Clear the cache agai, but this time for model_a
    cache.clear(model_id=model_a)

    # Check that all entries are cleared
    for i in range(n_entries):
        assert cache.get_entry(model_id=model_a, prompt=prompts[i], **params[i]) is None
    assert cache.count_entries() == {}


def test_cache_clearing_after(cache: CacheDB) -> None:
    """
    Test clearing the cache after a given datetime.
    """

    cache.clear()

    # Constants: model identifiers
    model_a = "model a"
    model_b = "model b"

    # We create 4 cache entries
    n_entries = 4

    # Constants: prompts
    prompts = [f"prompt {i}" for i in range(n_entries)]

    # Constants: responses
    responses = [[f"response {i}"] for i in range(n_entries)]

    # Constants: model parameters
    params: list[dict[str, str | int | float | bool]] = [
        dict(a_int=i, a_text=f"parameter {i}", a_float=i + 0.5, a_text_shared="shared")
        for i in range(n_entries)
    ]

    # Create a new cache database
    for i in range(0, n_entries - 1):
        cache.add_entry(
            model_id=model_a, prompt=prompts[i], responses=responses[i], **params[i]
        )

    time.sleep(1)
    timestamp_0 = now()
    time.sleep(1)
    # Create last cache entry
    cache.add_entry(
        model_id=model_a, prompt=prompts[-1], responses=responses[-1], **params[-1]
    )

    # Check that all entries are in the cache
    for i in range(n_entries):
        print(f"Checking entry {i}")
        assert (
            cache.get_entry(
                model_id=model_a,
                prompt=prompts[i],
                **params[i],
            )
            == responses[i]
        )

    # Clear the entry with the newest creation time
    cache.clear(created_after=timestamp_0)

    # Check that entry n is not in the cache, but the others are
    assert cache.get_entry(model_id=model_a, prompt=prompts[-1], **params[-1]) is None

    for i in range(0, n_entries - 1):
        assert (
            cache.get_entry(model_id=model_a, prompt=prompts[i], **params[i])
            == responses[i]
        )

    time.sleep(1)
    timestamp_1 = now()
    time.sleep(1)

    # Access entries 1, n-2
    for i in range(1, n_entries - 1):
        assert (
            cache.get_entry(model_id=model_a, prompt=prompts[i], **params[i])
            == responses[i]
        )

    # Clear all entries accessed after timestamp_1
    cache.clear(accessed_after=timestamp_1, model_id=model_a)

    # Confirm only entry 0 is in the cache
    assert (
        cache.get_entry(model_id=model_a, prompt=prompts[0], **params[0])
        == responses[0]
    )

    for i in range(1, n_entries):
        assert cache.get_entry(model_id=model_a, prompt=prompts[i], **params[i]) is None

    # Clear the cache
    cache.clear(model_id=model_b)

    # Confirm that we still have entry 1 in the cache
    assert (
        cache.get_entry(model_id=model_a, prompt=prompts[0], **params[0])
        == responses[0]
    )
    assert cache.count_entries() == {model_a: 1}

    # Clear the cache again, but this time for model_a
    cache.clear(model_id=model_a)

    # Check that all entries are cleared
    for i in range(n_entries):
        assert cache.get_entry(model_id=model_a, prompt=prompts[i], **params[i]) is None
    assert cache.count_entries() == {}


def test_cache_access_times(cache: CacheDB) -> None:
    """
    Test cache access times.
    """

    # Constants: model identifiers
    model_a = "model a"

    # Constants: prompts
    prompt_a = "prompt a"
    prompt_b = "prompt b"

    # Constants: responses
    response_single = "test_response"

    # Add a new single response
    t0 = now()
    cache.add_entry(
        model_id=model_a,
        prompt=prompt_a,
        responses=response_single,
    )
    t1 = now()

    # Get the single response
    assert cache.get_entry(model_id=model_a, prompt=prompt_a) == [response_single]

    # Get the access time of the cache entry
    access_times = cache.get_earliest_access_times()
    print(access_times)
    print(t0)
    print(t1)
    assert len(access_times) == 1
    assert t0 <= access_times[model_a]
    assert access_times[model_a] <= t1

    # Access the cache entry again
    t2 = now()
    cache.get_entry(model_id=model_a, prompt=prompt_a)
    t3 = now()

    # Get the access time of the cache entry
    access_times = cache.get_earliest_access_times()
    assert len(access_times) == 1
    assert t2 <= access_times[model_a] <= t3

    # The latest access time should be the same as the earliest access time because
    # there is only one cache entry
    assert cache.get_latest_access_times() == access_times

    # Get the creation time of the cache entry, this should be the same as the original
    # access time
    creation_times = cache.get_earliest_creation_times()
    assert len(creation_times) == 1
    assert t0 <= creation_times[model_a] <= t1

    # The latest creation time should be the same as the earliest creation time because
    # there is only one cache entry
    assert cache.get_latest_creation_times() == creation_times

    # Add a new single response
    t4 = now()
    cache.add_entry(
        model_id=model_a,
        prompt=prompt_b,
        responses=response_single,
    )
    t5 = now()

    # The earliest access time should be the same as before
    access_times = cache.get_earliest_access_times()
    assert len(access_times) == 1
    assert t2 <= access_times[model_a] <= t3

    # The latest access time should be newer
    access_times = cache.get_latest_access_times()
    assert len(access_times) == 1
    assert t4 <= access_times[model_a] <= t5

    # The earliest creation time should be the same as before
    creation_times = cache.get_earliest_creation_times()
    assert len(creation_times) == 1
    assert t0 <= creation_times[model_a] <= t1

    # The latest creation time should be newer
    creation_times = cache.get_latest_creation_times()
    assert len(creation_times) == 1
    assert t4 <= creation_times[model_a] <= t5


def test_invalid_values(cache: CacheDB) -> None:
    # Attempt to add an entry with an invalid parameter type
    with pytest.raises(
        TypeError,
        match=(
            r"^Model parameters must be strings, integers, floats, or booleans, but "
            r"got parameter invalid_param=\['invalid'\]$"
        ),
    ):
        cache.add_entry(
            model_id="model",
            prompt="prompt",
            responses="response",
            # a list is not a valid parameter type
            invalid_param=["invalid"],  # type: ignore[arg-type]
        )

    # Confirm that the cache is still empty
    assert cache.count_entries() == {}

    # Attempt to access an entry with an invalid parameter type
    with pytest.raises(
        TypeError,
        match=(
            r"^Model parameters must be strings, integers, floats, or booleans, but "
            r"got parameter invalid_param=\['invalid'\]$"
        ),
    ):
        cache.get_entry(
            model_id="model",
            prompt="prompt",
            # a list is not a valid parameter type
            invalid_param=["invalid"],  # type: ignore[arg-type]
        )


def test_parse_iso_utc() -> None:
    """
    Test the parse_iso_utc function.
    """
    # noinspection PyProtectedMember
    from artkit.model.cache._cache import _parse_iso_utc as parse_iso_utc

    # Test the function
    assert parse_iso_utc("2021-01-01T00:00:00") == datetime(
        2021, 1, 1, 0, 0, 0, tzinfo=timezone.utc
    )
    assert parse_iso_utc("2021-01-01T00:00:00Z") == datetime(
        2021, 1, 1, 0, 0, 0, tzinfo=timezone.utc
    )

    # Test the function with an invalid input
    with pytest.raises(ValueError):
        parse_iso_utc("2021-13-01T00:00:00")


#
# Auxiliary functions
#


def now() -> datetime:
    # Get the current time in UTC
    return datetime.now(tz=timezone.utc).replace(microsecond=0)



# -----------------------------------------------------------------------------
# © 2024 Boston Consulting Group. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
Bedrock LLM systems.
"""
from __future__ import annotations

import logging
from abc import ABCMeta
from typing import Any, TypeVar

from pytools.api import MissingClassMeta, appenddoc, inheritdoc, subsdoc

from ...base import HTTPXChatConnector

log = logging.getLogger(__name__)

__all__ = ["BaseBedrockChat"]

try:
    from botocore.auth import SigV4Auth
    from botocore.session import Session

except ImportError:

    class SigV4Auth(metaclass=MissingClassMeta, module="SigV4Auth"):  # type: ignore
        """Placeholder class for missing ``SigV4Auth`` class."""

    class Session(metaclass=MissingClassMeta, module="Session"):  # type: ignore
        """Placeholder class for missing ``Session`` class."""


try:
    from httpx import AsyncClient

except ImportError:

    class AsyncClient(metaclass=MissingClassMeta, module="httpx"):  # type: ignore
        """Placeholder class for missing ``AsyncClient`` class."""


__all__ = ["BaseBedrockChat"]

#
# Type variables
#

T_BaseBedrockChat = TypeVar("T_BaseBedrockChat", bound="BaseBedrockChat")

#
# Class declarations
#
log = logging.getLogger(__name__)


@inheritdoc(match="""[see superclass]""")
class BaseBedrockChat(HTTPXChatConnector, metaclass=ABCMeta):
    """
    Base class for Bedrock LLMs.
    """

    region: str | None

    @classmethod
    def get_default_api_key_env(cls) -> str:
        """[see superclass]"""
        return ""

    @subsdoc(
        # The pattern matches the row defining model_params, and move it to the end
        # of the docstring.
        pattern=r"(:param model_params: .*\n)((:?.|\n)*\S)(\n|\s)*",
        replacement=r"\2\1",
    )
    @appenddoc(to=HTTPXChatConnector.__init__)
    def __init__(
        self,
        *,
        model_id: str,
        api_key_env: str | None = None,
        initial_delay: float | None = None,
        exponential_base: float | None = None,
        jitter: bool | None = None,
        max_retries: int | None = None,
        system_prompt: str | None = None,
        httpx_client_kwargs: dict[str, Any] | None = None,
        region: str | None = None,
        **model_params: Any,
    ) -> None:
        """
        :param region: The specific AWS region to connect to.
        :raises CredentialsNotFoundError: if unable to find AWS Credentials.
        """
        super().__init__(
            model_id=model_id,
            api_key_env=api_key_env,
            initial_delay=initial_delay,
            exponential_base=exponential_base,
            jitter=jitter,
            max_retries=max_retries,
            system_prompt=system_prompt,
            httpx_client_kwargs=httpx_client_kwargs,
            **model_params,
        )

        self.region = region if region else "us-east-1"
        self.session = Session()
        self.credentials = self.session.get_credentials()
        self.auth = SigV4Auth(self.credentials, "bedrock", self.region)
        self.endpoint = f"https://bedrock-runtime.{self.region}.amazonaws.com/model/{self.model_id}/invoke"



import logging
import time
from io import StringIO

from pytest import LogCaptureFixture, fixture

import artkit.api as ak

log = logging.getLogger(__name__)

LOGS_TRUNCATED_MSG = "---- Logs truncated ----"


def test_log_throttling_exceeding_max_messages(
    log_throttling_handler: ak.LogThrottlingHandler,
    logging_test_msg: str,
    log_stream: StringIO,
) -> None:
    # Number of times we emit the log record
    n = log_throttling_handler.max_messages * 2
    for _ in range(n):
        log.warning(logging_test_msg)

    # Get the content from the IO stream
    log_stream_content = log_stream.getvalue().splitlines()

    # Check that the log stream contains the correct number of messages
    assert len(log_stream_content) == (log_throttling_handler.max_messages + 1)

    # Check that the log stream contains the correct messages
    assert log_stream_content == (
        [logging_test_msg for _ in range(log_throttling_handler.max_messages)]
        + [LOGS_TRUNCATED_MSG]
    )


def test_log_throttling_less_than_max_messages(
    log_throttling_handler: ak.LogThrottlingHandler,
    logging_test_msg: str,
    caplog: LogCaptureFixture,
) -> None:
    # Number of times we emit the log record
    n = int(log_throttling_handler.max_messages / 2)
    for _ in range(n):
        log.warning(logging_test_msg)
    same_messages = [
        record for record in caplog.records if record.msg == logging_test_msg
    ]
    assert len(same_messages) == n


def test_restart_logging_after_interval(
    log_throttling_handler: ak.LogThrottlingHandler,
    logging_test_msg: str,
    log_stream: StringIO,
) -> None:
    # Number of times we emit the log record
    n = log_throttling_handler.max_messages * 2
    for _ in range(n):
        log.warning(logging_test_msg)
    log_stream_content = log_stream.getvalue().splitlines()
    assert len(log_stream_content) == (log_throttling_handler.max_messages + 1)

    # Sleep for the interval
    time.sleep(log_throttling_handler.interval + 0.1)
    log.warning(logging_test_msg)
    # Get the content from the IO stream
    log_stream_content = log_stream.getvalue().splitlines()

    # Assert that the last message was sent normally again
    assert log_stream_content == (
        [logging_test_msg for _ in range(log_throttling_handler.max_messages)]
        + [LOGS_TRUNCATED_MSG]
        + [logging_test_msg]
    )


def test_that_changing_args_do_not_matter(
    log_throttling_handler: ak.LogThrottlingHandler,
    log_stream: StringIO,
) -> None:
    test_msg = "Test message %s"

    # Number of times we emit the log record
    n = log_throttling_handler.max_messages * 2
    for i in range(n):
        log.warning(test_msg, i)

    # Get the content from the IO stream
    log_stream_content = log_stream.getvalue().splitlines()

    # Check that the log stream contains the correct number of messages
    assert len(log_stream_content) == (log_throttling_handler.max_messages + 1)

    # Check that the log stream contains the correct messages
    assert log_stream_content == (
        [test_msg % i for i in range(log_throttling_handler.max_messages)]
        + [LOGS_TRUNCATED_MSG]
    )


#######################################################################################
#                                     FIXTURES                                        #
#######################################################################################


@fixture
def log_stream() -> StringIO:
    return StringIO()


@fixture
def log_throttling_handler(log_stream: StringIO) -> ak.LogThrottlingHandler:
    _throttling_handler = ak.LogThrottlingHandler(
        # We need to keep track of our own IO stream
        # since pytest capsys fixture struggles with logging
        # calls https://github.com/pytest-dev/pytest/issues/5997
        handler=logging.StreamHandler(log_stream),
        interval=0.5,
        max_messages=5,
    )
    log.addHandler(_throttling_handler)
    return _throttling_handler


@fixture
def logging_test_msg() -> str:
    return "This is a test log record for the log throttling handler."



import os
import shutil
from collections.abc import Iterator
from pathlib import Path
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from anthropic import RateLimitError

from artkit.model.llm import CachedChatModel
from artkit.model.llm.anthropic import AnthropicChat
from artkit.model.llm.base import ChatModel
from artkit.model.util import RateLimitException

anthropic = pytest.importorskip("anthropic")


@pytest.mark.asyncio
async def test_anthropic(anthropic_chat: AnthropicChat) -> None:
    # Mock anthropic Client
    with patch(
        "artkit.model.llm.anthropic._anthropic.AsyncAnthropic"
    ) as mock_get_client:
        # Mock anthropic Client response
        mock_response = AsyncMock(
            return_value=AsyncMock(content=[MagicMock(text="blue")], role="assistant")
        )

        # Set mock response as return value
        mock_get_client.return_value.messages.create = mock_response

        # Call mocked model
        messages = await anthropic_chat.get_response(
            message="What color is the sky? Please answer in one word."
        )
        assert "blue" in messages[0].lower()


@pytest.mark.asyncio
async def test_cached_anthropic(
    cached_anthropic: ChatModel,
) -> None:
    messages = await cached_anthropic.get_response(
        message="What color is the sky? Please answer in one word."
    )
    assert "blue" in messages[0].lower()


@pytest.mark.asyncio
async def test_anthropic_retry(
    anthropic_chat: AnthropicChat, caplog: pytest.LogCaptureFixture
) -> None:
    with patch(
        "artkit.model.llm.anthropic._anthropic.AsyncAnthropic"
    ) as MockClientSession:

        # Response mock
        response = AsyncMock()
        response.status_code = 429

        MockClientSession.return_value.messages.create.side_effect = RateLimitError(
            message="Rate Limit exceeded",
            response=response,
            body=AsyncMock(),
        )

        with pytest.raises(RateLimitException):
            await anthropic_chat.with_system_prompt(
                "Your job is to answer a quiz question with a single word, "
                "lowercase, with no punctuation"
            ).get_response(message="What color is the sky?")
        assert (
            MockClientSession.return_value.messages.create.call_count
            == anthropic_chat.max_retries
        )
    assert (
        len(
            [
                record
                for record in caplog.records
                if record.message.startswith("Rate limit exceeded")
            ]
        )
        == anthropic_chat.max_retries
    )


@pytest.fixture
def anthropic_chat() -> AnthropicChat:
    api_key_env = "TEST"
    os.environ[api_key_env] = "test"

    return AnthropicChat(
        max_tokens=10,
        temperature=1.0,
        model_id="claude-3-opus-20240229",
        api_key_env=api_key_env,
        initial_delay=0.1,
        exponential_base=1.5,
        max_retries=2,
    )


@pytest.fixture
def cached_anthropic(
    anthropic_chat: AnthropicChat, data_path: Path
) -> Iterator[ChatModel]:
    database = data_path / "_copy_anthropic.db"
    shutil.copyfile(data_path / "anthropic.db", database)

    yield CachedChatModel(model=anthropic_chat, database=database)

    os.remove(database)



# -----------------------------------------------------------------------------
# © 2024 Boston Consulting Group. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
Base implementation of diffusion model
"""

from __future__ import annotations

from abc import ABCMeta, abstractmethod

from ....util import Image

__all__ = [
    "DiffusionModel",
    "DiffusionModelConnector",
]

from typing import Any, Generic, TypeVar

from ...base import ConnectorMixin, GenAIModel
from ...util import retry_with_exponential_backoff

#
# Type variables
#

T_Client = TypeVar("T_Client")


#
# Classes
#


class DiffusionModel(GenAIModel, metaclass=ABCMeta):
    """An abstract diffusion model"""

    @abstractmethod
    async def text_to_image(
        self, text: str, **model_params: dict[str, Any]
    ) -> list[Image]:
        """
        Generate an image from text input

        :param text: the input text for the diffusion model
        :param model_params: additional parameters for the diffusion model
        :return: an :class:`.Image` object generated by the diffusion model
        """


class DiffusionModelConnector(
    DiffusionModel, ConnectorMixin[T_Client], Generic[T_Client], metaclass=ABCMeta
):
    """
    A diffusion model that connects to a client.
    """

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        # Apply the retry strategy to the image_to_text method
        cls.text_to_image = (  # type: ignore[method-assign]
            retry_with_exponential_backoff(cls.text_to_image)  # type: ignore
        )



from unittest.mock import AsyncMock, Mock, patch

import pytest
from httpx import HTTPStatusError, Response

from artkit.model.llm.base import ChatModel
from artkit.model.llm.huggingface import HuggingfaceURLChat
from artkit.model.util import RateLimitException

_ = pytest.importorskip("huggingface_hub")

#######################################################################################
#                                     Constants                                       #
#######################################################################################

EXAMPLE_URL = "http://huggingface.com"

#######################################################################################
#                                     UTILs                                           #
#######################################################################################


@pytest.mark.asyncio
async def test_huggingface_url_chat(
    hugging_face_url_chat: HuggingfaceURLChat,
) -> None:
    with patch(
        "artkit.model.llm.base.HTTPXChatConnector.get_client"
    ) as MockClientSession:

        # Mock the response object
        mock_post = Mock()
        mock_post.json = Mock(
            return_value={
                "choices": [{"message": {"role": "assistant", "content": "blue"}}]
            }
        )
        mock_post.text = AsyncMock()
        mock_post.return_value.status = 200

        # Set up the mock connection object
        mock_connection = AsyncMock()
        mock_connection.request.return_value = mock_post
        MockClientSession.return_value = mock_connection

        assert (
            await hugging_face_url_chat.with_system_prompt(
                "Your job is to answer a quiz question with a single word, "
                "lowercase, with no punctuation"
            ).get_response(message="What color is the sky?")
        )[0] == "blue"


@pytest.mark.asyncio
async def test_retry_huggingface_url_chat(
    hugging_face_url_chat: HuggingfaceURLChat,
    caplog: pytest.LogCaptureFixture,
) -> None:
    with patch(
        "artkit.model.llm.base.HTTPXChatConnector.get_client"
    ) as MockClientSession:

        # Set up the mock connection object
        mock_connection = AsyncMock()

        def f() -> None:
            err = HTTPStatusError(
                request=Mock(),
                response=Mock(spec=Response, status_code=429),
                message="Rate limit exceeded",
            )
            raise err

        mock_connection.request.return_value.raise_for_status = f
        MockClientSession.return_value = mock_connection

        # Test that request is being retried
        n_retries = hugging_face_url_chat.max_retries
        with pytest.raises(RateLimitException):
            assert (
                await hugging_face_url_chat.with_system_prompt(
                    "Your job is to answer a quiz question with a single word, "
                    "lowercase, with no punctuation"
                ).get_response(message="What color is the sky?")
            )[0] == "blue"
        assert mock_connection.request.call_count == n_retries

    # Test that warnings are given
    assert (
        len(
            [
                record
                for record in caplog.records
                if record.message.startswith("Rate limit exceeded")
            ]
        )
        == n_retries
    )


@pytest.mark.asyncio
async def test_unprocessable_huggingface_chat_aiohttp(
    hugging_face_url_chat: HuggingfaceURLChat,
) -> None:
    with patch(
        "artkit.model.llm.base.HTTPXChatConnector.get_client"
    ) as MockClientSession:

        # Set up the mock connection object
        mock_connection = AsyncMock()
        mock_connection.request.return_value.text.return_value = "Too many tokens"

        def f() -> None:
            err = HTTPStatusError(
                request=Mock(),
                response=Mock(spec=Response, status_code=422),
                message="Too many tokens",
            )
            raise err

        mock_connection.request.return_value.raise_for_status = f
        MockClientSession.return_value = mock_connection

        # Test that request fails
        with pytest.raises(ValueError):
            await hugging_face_url_chat.get_response(message="What color is the sky?")


def test_for_system_prompt(hugging_face_url_chat: HuggingfaceURLChat) -> None:
    chat = hugging_face_url_chat.with_system_prompt(
        "Your job is to answer a quiz question with a single word, "
        "lowercase, with no punctuation"
    )
    assert chat.system_prompt == (
        "Your job is to answer a quiz question with a single word, "
        "lowercase, with no punctuation"
    )


#######################################################################################
#                                     FIXTURES                                        #
#######################################################################################


@pytest.fixture
def hugging_face_url_chat(hf_token: str) -> ChatModel:
    return HuggingfaceURLChat(
        max_new_tokens=1,
        temperature=1.0,
        api_key_env=hf_token,
        model_id=EXAMPLE_URL,
        initial_delay=0.1,
        exponential_base=1.5,
        max_retries=2,
    )



# -----------------------------------------------------------------------------
# © 2024 Boston Consulting Group. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
Implementation of Titan Bedrock module.
"""

from __future__ import annotations

import logging
from typing import Any, TypeVar

from pytools.api import MissingClassMeta, inheritdoc

from ..history import AssistantMessage, ChatHistory, UserMessage
from .base import BaseBedrockChat

log = logging.getLogger(__name__)

__all__ = ["TitanBedrockChat"]

try:
    from botocore.awsrequest import AWSRequest
except ImportError:

    class AWSRequest(metaclass=MissingClassMeta, module="AWSRequest"):  # type: ignore
        """Placeholder class for missing ``AWSRequest`` class."""


try:
    from httpx import Response

except ImportError:

    class Response(metaclass=MissingClassMeta, module="httpx"):  # type: ignore
        """Placeholder class for missing ``Response`` class."""


#
# Constants
#
TITAN_USER_ROLE = "User: "
TITAN_BOT_ROLE = "Bot: "
TITAN_EOT_TOKEN = "\n"
TITAN_ROLE_MAPPING = {
    UserMessage.ROLE: TITAN_USER_ROLE,
    AssistantMessage.ROLE: TITAN_BOT_ROLE,
}

#
# Type variables
#

T_TitanBedrockChat = TypeVar("T_TitanBedrockChat", bound="TitanBedrockChat")

#
# Class declarations
#


@inheritdoc(match="""[see superclass]""")
class TitanBedrockChat(BaseBedrockChat):
    """
    Implementation of a BaseBedrockChat with Titan LLM as the Foundational Model.
    """

    def build_request_arguments(
        self,
        message: str,
        *,
        history: ChatHistory | None = None,
        **model_params: dict[str, Any],
    ) -> dict[str, Any]:
        """[see superclass]"""
        if self.system_prompt is not None:
            raise ValueError(
                "Titan does not support system prompts. Please remove the system_prompt parameter."
            )
        text = ""
        if history is not None:
            for historic_message in history.messages:
                role = historic_message.role
                message_txt = historic_message.text
                text += role + message_txt + TITAN_EOT_TOKEN

        text += TITAN_USER_ROLE + message + TITAN_EOT_TOKEN
        request = AWSRequest(
            method="POST",
            url=self.endpoint,
            data=message,
            headers={"content-type": "application/json"},
        )
        self.auth.add_auth(request)
        prepped_request = request.prepare()
        headers = dict(prepped_request.headers.items())
        method = "POST"
        return dict(
            method=method,
            url=prepped_request.url,
            headers=headers,
            data=prepped_request.body,
        )

    def parse_httpx_response(self, response: Response) -> list[str]:
        """[see superclass]"""
        response_body = response.json()
        formatted_responses = []
        for result in response_body["results"]:
            output_text = result["outputText"]
            # Example: stripping unnecessary whitespace and adding a line break
            formatted_text = output_text.strip()
            formatted_text = formatted_text.replace(
                "\n", "\n\n"
            )  # Add extra line breaks for readability

            formatted_responses.append(formatted_text)

        return list(formatted_responses)



# -----------------------------------------------------------------------------
# © 2024 Boston Consulting Group. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
Abstract base class to represent connecting to a custom endpoint.
"""
from __future__ import annotations

import asyncio
import logging
from abc import ABCMeta, abstractmethod
from typing import Any
from weakref import WeakKeyDictionary

from pytools.api import MissingClassMeta, appenddoc, inheritdoc, subsdoc

from ...base import ClientWrapper
from ...util import RateLimitException
from ..history import ChatHistory
from ._llm import ChatModelConnector

try:
    from httpx import AsyncClient, HTTPStatusError, Response

except ImportError:

    class AsyncClient(metaclass=MissingClassMeta, module="httpx"):  # type: ignore
        """Placeholder class for missing ``AsyncClient`` class."""

    class Response(metaclass=MissingClassMeta, module="httpx"):  # type: ignore
        """Placeholder class for missing ``Response`` class."""

    class HTTPStatusError(metaclass=MissingClassMeta, module="httpx"):  # type: ignore
        """Placeholder class for missing ``HTTPStatusError`` class."""


log = logging.getLogger(__name__)

__all__ = ["HTTPXChatConnector"]

#
# Class Definitions
#


@inheritdoc(match="""[see superclass]""")
class HTTPXChatConnector(ChatModelConnector[AsyncClient], metaclass=ABCMeta):
    """
    ABC that represents connecting to a custom endpoint.
    """

    @subsdoc(
        # The pattern matches the row defining model_params, and move it to the end
        # of the docstring.
        pattern=r"(:param model_params: .*\n)((:?.|\n)*\S)(\n|\s)*",
        replacement=r"\2\1",
    )
    @appenddoc(to=ChatModelConnector.__init__)
    def __init__(
        self,
        *,
        model_id: str,
        api_key_env: str | None = None,
        initial_delay: float | None = None,
        exponential_base: float | None = None,
        jitter: bool | None = None,
        max_retries: int | None = None,
        system_prompt: str | None = None,
        httpx_client_kwargs: dict[str, Any] | None = None,
        **model_params: Any,
    ) -> None:
        """
        :param httpx_client_kwargs: the kwargs to use for making
         the :class:`httpx.AsyncClient` (optional)
        """
        super().__init__(
            model_id=model_id,
            api_key_env=api_key_env,
            initial_delay=initial_delay,
            exponential_base=exponential_base,
            jitter=jitter,
            max_retries=max_retries,
            system_prompt=system_prompt,
            **model_params,
        )

        if httpx_client_kwargs is None:
            httpx_client_kwargs = {}
        self.httpx_client_kwargs = httpx_client_kwargs

        # Set up caching for clients based on event loop
        # Use weak references to enable garbage collection
        self._async_clients: WeakKeyDictionary[
            asyncio.AbstractEventLoop, ClientWrapper[AsyncClient]
        ] = WeakKeyDictionary()

    def _make_client(self) -> AsyncClient:
        return AsyncClient(**self.httpx_client_kwargs)

    def get_client(self) -> AsyncClient:
        """
        Get the client to use for making requests.
        Allows us to create a new async client every time we're in a new event loop.
        Internally, we can cache the client using a weak key dictionary that maps event
        loop objects to async clients

        :return: The client to use for making requests.
        """
        # get event loop from current thread
        current_loop = asyncio.get_running_loop()
        if current_loop in self._async_clients:
            return self._async_clients[current_loop].client
        else:
            self._async_clients[current_loop] = ClientWrapper(self._make_client())
            return self._async_clients[current_loop].client

    @abstractmethod
    def build_request_arguments(
        self,
        message: str,
        *,
        history: ChatHistory | None = None,
        **model_params: dict[str, Any],
    ) -> dict[str, Any]:
        """
        This method is responsible for formatting the input to the LLM chat system.
        For argument options see :class:`httpx.AsyncClient.request`.

        :param message: The input message to format.
        :param history: The chat history preceding the message.
        :param model_params: Additional parameters for the chat system.

        :return: The necessary httpx request arguments.
        """

    @abstractmethod
    def parse_httpx_response(self, response: Response) -> list[str]:
        """
        This method is responsible for formatting the :class:`httpx.Response` after
        having made the request.

        :param response: The response from the endpoint.
        :return: A list of formatted response strings.
        """

    async def get_response(
        self,
        message: str,
        *,
        history: ChatHistory | None = None,
        **model_params: dict[str, Any],
    ) -> list[str]:
        """[see superclass]"""
        client = self.get_client()
        response = await client.request(
            **self.build_request_arguments(
                message=message, history=history, **model_params
            )
        )
        try:
            response.raise_for_status()
        except HTTPStatusError as e:
            if e.response.status_code == 429:
                raise RateLimitException(
                    "Rate limit exceeded. Please try again later."
                ) from e
            elif e.response.status_code == 422:
                raise ValueError(
                    f"Invalid request. Please check the request parameters. {e.response.text}"
                ) from e
            raise
        return self.parse_httpx_response(response=response)



# -----------------------------------------------------------------------------
# © 2024 Boston Consulting Group. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
Implementation of the `coerce_json_format` function, which uses an LLM to coerce a
JSON-like string into the specified dictionary format. The function is used to
validate the JSON format of the result and coerce it into the expected format.
"""

import logging
from collections.abc import Callable
from json import JSONDecodeError
from json import loads as json_loads
from typing import Any

from artkit.model.llm.base import ChatModel

log = logging.getLogger(__name__)

__all__ = [
    "parse_json_autofix",
]


#
# Constants
#

# System prompt to fix JSON syntax errors
_JSON_REPAIR_SYNTAX_SYSTEM_PROMPT = """\
You must correct the syntax in a JSON string provided by the user. \
Take your time to ensure the syntax is correct. \
Return only the corrected JSON, with no additional formatting or context."""

# System prompt to fix JSON validation error
_JSON_REPAIR_CONTENT_SYSTEM_PROMPT = """\
You must correct the following ERROR in a JSON string provided by the user. \
Take your time to ensure the resulting JSON corrects the error and has correct syntax. \
Return only the corrected JSON, with no additional formatting or context.

ERROR: {error}"""


#
# Functions
#


async def parse_json_autofix(
    json: str,
    *,
    model: ChatModel,
    validator: Callable[[Any], str | None] | None = None,
) -> Any:
    """
    Convert a string in JSON format to a Python object, using the given chat model to
    fix the JSON format if parsing fails.

    An optional validator function can be provided to validate the resulting object,
    triggering another attempt to fix the response if the object is invalid. The
    function must return a string with an error description if the object is invalid,
    and ``None`` if the object is valid. The error description will be passed to the
    chat model as part of the instructions to fix the JSON format.

    :param json: the JSON string to convert to a dictionary
    :param model: the chat model used to repair the JSON format, if necessary
    :param validator: an optional function to validate the resulting object, returning
        ``True`` if the object is valid, and ``False`` otherwise
    :return: the resulting object
    """

    # Attempt to load the JSON string
    try:
        parsed = json_loads(json)
    except JSONDecodeError:
        # Parsing failed
        log.warning(f"Attempting to fix malformed JSON:\n{json}")
        # Ask the model to fix the JSON syntax; we only use the first response in case
        # there are multiple responses
        json_fixed = (
            await model.with_system_prompt(
                _JSON_REPAIR_SYNTAX_SYSTEM_PROMPT
            ).get_response(message=json)
        )[0]
        try:
            parsed = json_loads(json_fixed)
        except JSONDecodeError as e:
            # Parsing failed again: give up
            raise ValueError(
                f"Failed to fix malformed JSON with error {str(e)!r}:\n{json_fixed}"
            )

        # Parsing succeeded: use the fixed JSON
        json = json_fixed

    # Validate the resulting object
    if validator is None:
        # No validation function was provided
        return parsed

    # We have a validation function: check if the object is valid
    error = validator(parsed)
    if error is None:
        # Validation was successful
        return parsed

    # The object is invalid
    log.warning(
        f"Attempting to fix failed validation with error {error!r} for:\n{json}"
    )
    # Ask the model to fix the JSON content using the error message for guidance
    json_fixed = (
        await model.with_system_prompt(
            _JSON_REPAIR_CONTENT_SYSTEM_PROMPT.format(
                # We insert the error message into the system prompts
                error=error
            )
        ).get_response(message=json)
    )[0]

    try:
        # Attempt to load the fixed JSON
        parsed = json_loads(json_fixed)

    except JSONDecodeError as e:
        # Parsing failed
        log.warning(
            f"Malformed JSON with error {str(e)!r} after attempting to fix "
            f"validation error:\n{json_fixed}"
        )

    else:
        # Re-validate the fixed object
        new_error = validator(parsed)
        if new_error is None:
            # Fixed JSON is valid: return it
            return parsed

        # The fixed JSON is still invalid
        log.warning(
            f"New validation error {new_error!r} after attempting to fix the "
            f"previous error:\n{json_fixed}"
        )

    # Parsing failed: give up
    raise ValueError(f"Failed to fix JSON after validation error {error!r}:\n{json}")



# -----------------------------------------------------------------------------
# © 2024 Boston Consulting Group. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
Implementation of the Anthropic llm module.
"""

import logging
from collections.abc import Iterator
from contextlib import AsyncExitStack
from typing import Any, TypeVar

from pytools.api import MissingClassMeta, inheritdoc

from ...util import RateLimitException
from ..base import ChatModelConnector
from ..history import ChatHistory

log = logging.getLogger(__name__)

try:  # pragma: no cover
    from anthropic import AsyncAnthropic, RateLimitError
    from anthropic.types import Message
except ImportError:

    class AsyncAnthropic(metaclass=MissingClassMeta, module="anthropic"):  # type: ignore
        """
        Placeholder class for missing ``AsyncAnthropic`` class.
        """

        def __init__(self, *args: Any, **kwargs: Any) -> None:
            pass

    class Message(metaclass=MissingClassMeta, module="anthropic.types"):  # type: ignore
        """
        Placeholder class for missing ``Message`` class.
        """

    class RateLimitError(metaclass=MissingClassMeta, module="anthropic"):  # type: ignore
        """
        Placeholder class for missing ``RateLimitError`` class.
        """


__all__ = [
    "AnthropicChat",
]


#
# Type variables
#

T_AnthropicLLM = TypeVar("T_AnthropicLLM", bound="AnthropicChat")

#
# Class declarations
#


@inheritdoc(match="""[see superclass]""")
class AnthropicChat(ChatModelConnector[AsyncAnthropic]):
    """
    An asynchronous Anthropic LLM.
    """

    @classmethod
    def get_default_api_key_env(cls) -> str:
        """[see superclass]"""
        return "ANTHROPIC_API_KEY"

    async def get_response(
        self,
        message: str,
        *,
        history: ChatHistory | None = None,
        **model_params: dict[str, Any],
    ) -> list[str]:
        """[see superclass]"""
        async with AsyncExitStack():

            try:
                message = await self.get_client().messages.create(
                    messages=list(
                        self._messages_to_anthropic_format(message, history=history)
                    ),
                    model=self.model_id,
                    **{
                        **(
                            {"system": self.system_prompt} if self.system_prompt else {}
                        ),
                        **self.get_model_params(),
                        **model_params,
                    },
                )
            except RateLimitError as e:
                raise RateLimitException(
                    "Rate limit exceeded. Please try again later."
                ) from e

        return list(self._responses_from_message(message))

    def _make_client(self) -> AsyncAnthropic:  # pragma: no cover
        return AsyncAnthropic(api_key=self.get_api_key())

    @staticmethod
    def _messages_to_anthropic_format(
        user_message: str, *, history: ChatHistory | None = None
    ) -> Iterator[dict[str, str]]:
        """
        Get the messages to send to the Anthropic LLM, based on the given user prompt
        and chat history, and the system prompt for this LLM.

        :param user_message: the user prompt to send to the Anthropic LLM
        :param history: the chat history to include in the messages (optional)
        :return: the messages object, in the format expected by the Anthropic API
        """
        if history is not None:
            for message in history.messages:
                yield {"role": message.role, "content": message.text}

        yield {"role": "user", "content": user_message}

    @staticmethod
    def _responses_from_message(message: Message) -> Iterator[str]:
        """
        Get the response from the given chat completion.

        :param message: the chat completion to process
        :return: the alternate responses from the chat completion
        """

        for content in message.content:
            if message.role != "assistant":
                log.warning("Expected only assistant messages, but got " f"{content!r}")
            yield content.text



# -----------------------------------------------------------------------------
# © 2024 Boston Consulting Group. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
Implementation of the ``CachedDiffusionModel`` class.
"""

from typing import Any, Generic, TypeVar

from pytools.api import inheritdoc

from ...util import Image
from ..base import CachedGenAIModel
from .base import DiffusionModel

__all__ = [
    "CachedDiffusionModel",
]

#
# Type variables
#
# Naming convention used here:
# _ret for covariant type variables used in return positions
# _arg for contravariant type variables used in argument positions


T_DiffusionModel_ret = TypeVar(
    "T_DiffusionModel_ret", bound=DiffusionModel, covariant=True
)


#
# Classes
#


@inheritdoc(match="""[see superclass]""")
class CachedDiffusionModel(
    CachedGenAIModel[T_DiffusionModel_ret, list[str]],
    DiffusionModel,
    Generic[T_DiffusionModel_ret],
):
    """
    A wrapper around a diffusion model to cache results
    """

    async def text_to_image(
        self, text: str, **model_params: dict[str, Any]
    ) -> list[Image]:
        """[see superclass]"""

        # Include model parameters in cache key
        model_params_merged = {
            "_type": "diffusion",
            **self.get_model_params(),
            **model_params,
        }

        # Try to get images from cache
        response: list[str] | None = self._get(prompt=text, **model_params_merged)
        if response is None:
            # Get Image objects from model and cache b64 string
            images: list[Image] = await self.model.text_to_image(
                text=text, **model_params
            )
            self._put(
                prompt=text,
                responses=[
                    img.data.decode(
                        # decode bytes to a string, without changing the data
                        "latin1"
                    )
                    for img in images
                ],
                **model_params_merged,
            )
            return images
        else:
            # Decode and return the images from the cache
            return [
                Image(
                    # decode a string back to bytes
                    data=d.encode("latin1")
                )
                for d in response
            ]



# -----------------------------------------------------------------------------
# © 2024 Boston Consulting Group. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
Implementation of Titan Bedrock module.
"""

from __future__ import annotations

import base64
import logging
from typing import Any, TypeVar

from pytools.api import inheritdoc

from ....util import Image
from .base import BaseBedrockDiffusionModel

log = logging.getLogger(__name__)

__all__ = ["TitanBedrockDiffusionModel"]

#
# Type variables
#

T_TitanBedrockDiffusionModel = TypeVar(
    "T_TitanBedrockDiffusionModel", bound="TitanBedrockDiffusionModel"
)

#
# Class declarations
#


@inheritdoc(match="""[see superclass]""")
class TitanBedrockDiffusionModel(BaseBedrockDiffusionModel):
    """
    Implementation of a BaseBedrockDiffusionModel with Titan.
    """

    async def text_to_image(
        self, text: str, **model_params: dict[str, Any]
    ) -> list[Image]:
        """[see superclass]"""
        response_body = await self.get_response_from_model(text)
        images = []
        for img_data in response_body.get("images", []):
            base64_bytes = img_data.encode("ascii")
            img_bytes = base64.b64decode(base64_bytes)
            image = Image(data=img_bytes)
            images.append(image)
        return images



"""
Configuration file for the Sphinx documentation builder.

For a full list of options see the documentation:
https://www.sphinx-doc.org/en/master/usage/configuration.html
"""

import logging
import os
import shutil
import sys
from typing import Any

from sphinx.application import Sphinx

from pytools.sphinx.util import (
    Replace3rdPartyDoc,
    ResolveTypeVariables,
    UpdateForwardReferences,
)

logging.basicConfig(level=logging.INFO)
_log = logging.getLogger(name=__name__)

# this is the directory that contains all required repos
_dir_conf_base = os.path.dirname(os.path.abspath(__file__))
_dir_sphinx = os.path.dirname(_dir_conf_base)
_dir_src = os.path.join(os.path.dirname(_dir_sphinx), "src")


# noinspection PyShadowingNames
def set_config(
    globals_: dict[str, Any],
    *,
    project: str,
    html_logo: str | None = None,
    intersphinx_mapping: dict[str, tuple[str, str | None]] | None = None,
) -> None:
    """
    Add required modules to the python path, and set custom configuration options
    """

    from make_util import get_package_version

    globals_["project"] = project
    globals_["version"] = str(
        get_package_version(package_path=os.path.join(_dir_src, project))
    )

    globals_["html_show_sourcelink"] = False

    if html_logo:
        globals_["html_logo"] = html_logo
        globals_["latex_logo"] = html_logo

    if intersphinx_mapping:
        _update_intersphinx_mapping(intersphinx_mapping)

    module_path = _dir_src
    if module_path not in sys.path:
        # noinspection PyUnboundLocalVariable
        sys.path.insert(0, module_path)
        _log.info(f"added '{module_path}' to python paths")

    # Update global variables
    globals_.update(
        (k, v) for k, v in globals().items() if not (k.startswith("_") or k in globals_)
    )


def _update_intersphinx_mapping(mapping: dict[str, tuple[str, str | None]]) -> None:
    global intersphinx_mapping
    intersphinx_mapping.update(mapping)


_log.info(f"sys.path = {sys.path}")

# -- Project information -----------------------------------------------------

# noinspection PyShadowingBuiltins
copyright = "2024, Boston Consulting Group (BCG)"
author = "BCG X Responsible AI Team"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "nbsphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.mathjax",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx_autodoc_typehints",
]

# -- Options for autodoc / autosummary -------------------------------------------------

# prevent numpydoc from interfering with autosummary
numpydoc_show_class_members = False

# generate autosummary even if no references
autosummary_generate = False

# always overwrite generated auto summaries with newly generated versions
autosummary_generate_overwrite = True

autosummary_imported_members = True

autodoc_default_options = {
    "ignore-module-all": False,
    "inherited-members": True,
    "imported-members": False,
    "show-inheritance": False,
    "member-order": "groupwise",
}

nbsphinx_allow_errors = True

# add intersphinx mapping
intersphinx_mapping: dict[str, tuple[str, str | None]] = {
    "fluxus": ("https://bcg-x-official.github.io/fluxus", None),
    "matplotlib": ("https://matplotlib.org/stable", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable", None),
    "pd": ("https://pandas.pydata.org/pandas-docs/stable", None),
    "python": ("https://docs.python.org/3.7", None),
}

intersphinx_collapsible_submodules = {
    "pandas.core.frame": "pandas",
    "pandas.core.series": "pandas",
    "pandas.core.indexes.base": "pandas",
    "pandas.core.indexes.multi": "pandas",
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["*/.ipynb_checkpoints/*", "**/supplementary/**"]

# -- Options for sphinx_autodoc_typehints ----------------------------------------------
set_type_checking_flag = False
typehints_fully_qualified = False
always_document_param_types = False
autodoc_typehints_format = "short"

# -- Options for Math output -----------------------------------------------------------

imgmath_image_format = "svg"
imgmath_use_preview = True

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

html_theme_options = {
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/BCG-X-Official/artkit",
            "icon": "fab fa-github",
        },
    ],
    "collapse_navigation": True,
    "show_nav_level": 1,
    "show_toc_level": 2,
    "navigation_depth": 6,
    # Omit the `theme-switcher` since we don't support dark mode
    # (as of pydata theme 0.9):
    "navbar_end": ["navbar-icon-links"],
    # "secondary_sidebar_items": {"path/to/page": [],},
    "secondary_sidebar_items": ["page-toc", "sourcelink"],
    "show_prev_next": False,
}

html_sidebars: dict[str, list[str]] = {
    "_generated/home": [],
    "faq": [],
    "_generated/release_notes": [],
}

# Set the default mode to "light" instead of "auto" (as of pydata theme 0.9):
html_context = {"default_mode": "light"}


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["../make/_static"]
html_logo = os.path.join("_static", "bcgx_logo.jpg")

# Class documentation to include docstrings both global to the class, and from __init__
autoclass_content = "both"

# -- End of options section ------------------------------------------------------------


def setup(app: Sphinx) -> None:
    """
    Add event handlers to the Sphinx application object.

    :param app: the Sphinx application object
    """

    _connect_callbacks(app=app)

    _add_custom_css_and_js(app=app)


def _connect_callbacks(app: Sphinx) -> None:
    from pytools.sphinx.util import (
        AddInheritance,
        CollapseModulePathsInDocstring,
        CollapseModulePathsInSignature,
        CollapseModulePathsInXRef,
        RenamePrivateArguments,
        SkipIndirectImports,
        TrackCurrentClass,
    )

    UpdateForwardReferences().connect(app=app)

    AddInheritance(collapsible_submodules=intersphinx_collapsible_submodules).connect(
        app=app
    )

    CollapseModulePathsInDocstring(
        collapsible_submodules=intersphinx_collapsible_submodules
    ).connect(app=app, priority=100000)

    CollapseModulePathsInSignature(
        collapsible_submodules=intersphinx_collapsible_submodules
    ).connect(app=app, priority=100000)

    SkipIndirectImports().connect(app=app)

    Replace3rdPartyDoc().connect(app=app)

    TrackCurrentClass().connect(app=app)

    ResolveTypeVariables().connect(app=app)

    CollapseModulePathsInXRef(
        collapsible_submodules=intersphinx_collapsible_submodules
    ).connect(app=app)

    RenamePrivateArguments().connect(app=app)


def _add_custom_css_and_js(app: Sphinx) -> None:
    # add custom css and js files, and copy them to the build/html/_static folder
    css_rel_paths = [os.path.join("css", "bcgx.css")]
    js_rel_paths = [os.path.join("js", "bcgx.js"), os.path.join("js", "versions.js")]

    for css in css_rel_paths:
        app.add_css_file(filename=css)

    for js in js_rel_paths:
        app.add_js_file(filename=js)

    src_root = os.path.join(_dir_sphinx, "make", "_static")
    dst_html = os.path.join(_dir_sphinx, "build", "html", "_static")
    for rel_path in css_rel_paths + js_rel_paths:
        dst_dir = os.path.join(dst_html, os.path.dirname(rel_path))
        os.makedirs(dst_dir, exist_ok=True)
        src_file = os.path.join(src_root, rel_path)
        _log.info(f"copying {src_file} to {dst_dir}")
        shutil.copy(src=src_file, dst=dst_dir)
